(** TODO: Maybe some nice parallel examples directly with VCG here?
  currently, all more complex stuff is done via Sepref, and there's 
  only the minimum required VCG setup!
*)
theory Par_Array_Examples
imports "../sepref/IICF/Impl/IICF_Array"
begin
  declare [[llc_compile_par_call=true]]

  context fixes f :: "'a \<Rightarrow> 'a" begin

  fun abs_map where
    "abs_map xs = (
      if length xs \<le> 1 then map f xs 
      else let
        xs\<^sub>1 = take (length xs div 2) xs;
        xs\<^sub>2 = drop (length xs div 2) xs;
        xs\<^sub>1 = abs_map xs\<^sub>1;
        xs\<^sub>2 = abs_map xs\<^sub>2
      in
        xs\<^sub>1@xs\<^sub>2
      )"
    
    lemmas [simp del] = abs_map.simps  

      
    lemma "abs_map xs = map f xs"
      apply (induction xs rule: abs_map.induct)
      apply (subst abs_map.simps)
      apply (auto split: if_splits)
      by (metis atd_lem map_append)
    
    
    definition "map1 \<equiv> REC (\<lambda>map1 (xs,n). doN {
      ASSERT (n = length xs);
      if length xs = 0 then RETURN xs
      else if length xs = 1 then doN {
        x \<leftarrow> mop_list_get xs 0;
        xs \<leftarrow> mop_list_set xs 0 (f x);
        RETURN xs
      } else doN {
        let m = n div 2;
        (_,xs) \<leftarrow> WITH_SPLIT m xs (\<lambda>xs\<^sub>1 xs\<^sub>2. do {
          (xs\<^sub>1,xs\<^sub>2) \<leftarrow> nres_par map1 map1 (xs\<^sub>1,m) (xs\<^sub>2,n-m);
          RETURN ((),(xs\<^sub>1,xs\<^sub>2))
        });
        RETURN xs
      }
      })"  
      
          
  end  

    
    
  context
    includes monad_syntax_M
  begin
  
  definition par_incr :: "8 word ptr \<times> 64 word \<Rightarrow> 8 word ptr llM"
  where [llvm_code]: "par_incr \<equiv> MMonad.REC (\<lambda>par_incr (p,n).
    if n=1 then doM {
      x \<leftarrow> array_nth p (0::64 word);
      x \<leftarrow> ll_udiv x 2;
      array_upd p (0::64 word) x;
      return p
    } else doM {
      i \<leftarrow> ll_sdiv n 2;
      (p\<^sub>1,p\<^sub>2) \<leftarrow> ars_split i p;
      llc_par par_incr par_incr (p\<^sub>1,i) (p\<^sub>2,n-i);
      ars_join p\<^sub>1 p\<^sub>2;
      return p
    })"
  
  end  
  
  export_llvm par_incr
  
  schematic_goal par_incr_unfold: "par_incr = ?foo"
    unfolding par_incr_def
    apply (subst MMonad.REC_unfold)
    apply pf_mono_prover
    unfolding par_incr_def[symmetric]
    ..
  
    
  lemma "llvm_htriple 
    (raw_array_slice_assn xs xsi ** snat_assn n ni ** \<up>(n=length xs)) 
    (par_incr (xsi,ni))
    (\<lambda>_. raw_array_slice_assn (abs_map (\<lambda>x. x div 2) xs) xsi)
    "
  proof (induction xs rule: abs_map.induct[where f="(\<lambda>x. x div 2)"])
    case (1 xs)
    show ?case
      apply (subst par_incr_unfold)
      apply (rule htriple_pure_preI, ((determ \<open>drule pure_part_split_conj|erule conjE\<close>)+)?)
      apply vcg_monadify
      
      find_theorems pure_part LLVM_DS_NArray.array_slice_assn
      
      apply vcg
      apply vcg_rl
      apply vcg_try_solve
    
    
  qed
    
    
    unfolding par_incr_def
    apply (rule REC_rule[where V="measure nat" and P="\<lambda>t ()"])
    apply vcg
    apply pf_mono_prover
  
  
  
  
  
  
  
  
  lemma REC_rule: (* Needs annotation to be usable in automated setting! *)
    assumes WF: "wf V"
    assumes MONO: "M_mono_body B"
    assumes STEP: "\<And>D t x. \<lbrakk>\<And>t' x. (t',t)\<in>V \<Longrightarrow> htriple (P t' x ** F) (D x) (Q x) \<rbrakk> \<Longrightarrow> htriple (P t x ** F) (B D x) (Q x)"
    assumes INIT: "P' \<turnstile> P t x ** F"
    shows "htriple P' (MMonad.REC B x) (Q x)"
  proof -  
  
    have "htriple (P t x ** F) (MMonad.REC B x) (Q x)"
      using WF
      apply (induction t arbitrary: x)
      apply (subst MMonad.REC_unfold[OF MONO])
      by (rule STEP)
    thus ?thesis
      apply (rule cons_rule)
      subgoal using INIT unfolding entails_def by simp
      subgoal .
      .
  qed



  declare [[llc_compile_par_call=true]]

  find_theorems raw_array_slice_assn

  context
    includes monad_syntax_M
  begin

  definition [llvm_code]: ""
  
  
  
  definition [llvm_code]: "par_incr (p::8 word ptr) (n::64 word) \<equiv> MMonad.REC (\<lambda>par_incr (p,n).
    if n=1 then doM {
      x \<leftarrow> array_nth p (0::64 word);
      x \<leftarrow> ll_udiv x 2;
      array_upd p (0::64 word) x;
      return p
    } else doM {
      i \<leftarrow> ll_sdiv n 2;
      (p\<^sub>1,p\<^sub>2) \<leftarrow> ars_split i p;
      llc_par par_incr par_incr (p\<^sub>1,i) (p\<^sub>2,n-i);
      ars_join p\<^sub>1 p\<^sub>2;
      return p
    }) (p,n)
  "

  
  export_llvm par_incr
  
  lemma
    shows ""
  
  
  
  lemma "llvm_htriple 
    (raw_array_slice_assn xs xsi ** snat_assn n ni ** \<up>(n=length xs)) 
    (par_incr xsi ni)
    (\<lambda>_. raw_array_slice_assn (map (\<lambda>x. x div 2) xs) xsi)
    "
    unfolding par_incr_def
    apply (rule REC_rule[where V="measure nat" and P="\<lambda>t ()"])
    apply vcg
    apply pf_mono_prover
    
  
    find_theorems MMonad.REC 
    
  
  end



end
