theory Ndet_Monad
imports Basic_Imports RefineG_Domain (*Refine_Mono_Prover*) (*Frame_Infer*)
begin


  (* Very simple memory model *)

  datatype 'v cell = FRESH | FREED | VALID 'v

  class infinite =
    assumes infinite_UNIV[simp,intro!]: "\<not> finite (UNIV :: 'a set)"
  
    
  typedef ('a::infinite,'v) memory = "{ f :: 'a \<Rightarrow> 'v cell . infinite {a. f a = FRESH} }" 
    by (auto intro: exI[where x="\<lambda>_. FRESH"])

  setup_lifting type_definition_memory  

  






  


  text \<open>Handy tag for simplifier-based synthesis. 
    Simplify \<open>EQ_TAG t ?b\<close>, and then use \<open>EQI\<close> to obtain
    theorem relating \<open>t\<close> with its simplified version.
  \<close>
  definition "EQ_TAG a b \<equiv> a=b"  
  lemma EQD: "EQ_TAG a b \<Longrightarrow> a=b" by (simp add: EQ_TAG_def)
  lemma EQI: "EQ_TAG a a"  by (simp add: EQ_TAG_def)



  subsection \<open>Complete Lattices Additions\<close>

  locale complete_lattice_from_Sup = 
    fixes fSup :: "'a set \<Rightarrow> 'a"
    assumes fSup_singleton[simp]: "fSup {x} = x"
    assumes fSup_finite_join: "fSup { fSup X, fSup Y } = fSup (X\<union>Y)"
    assumes fSup_add_less: "x\<in>X \<Longrightarrow> \<forall>y\<in>X. fSup {y, x} = x \<Longrightarrow> fSup X = x"
    
  begin
    definition "fsup a b = fSup {a,b}"
    definition "fless_eq a b \<equiv> fsup a b = b"
    definition "fInf X \<equiv> fSup {b. \<forall>a\<in>X. fless_eq b a}"
    definition "finf a b = fInf {a,b}"
    definition "fless a b \<longleftrightarrow> fless_eq a b \<and> a\<noteq>b"
    definition "ftop = fSup UNIV"
    definition "fbot = fSup {}"
    
    lemma fSup_join_simps:
      "fSup {x, fSup Y} = fSup (insert x Y)"
      "fSup {fSup X, y} = fSup (insert y X)"
      apply (metis fSup_finite_join fSup_singleton insert_is_Un)
      by (metis fSup_finite_join fSup_singleton insert_commute insert_is_Un)
    
    lemma is_complete_lattice: "class.complete_lattice fInf fSup finf fless_eq fless fsup fbot ftop"
      apply standard
      unfolding fsup_def fless_eq_def fInf_def finf_def fless_def ftop_def fbot_def 
      apply (auto simp: insert_commute insert_absorb fSup_join_simps intro: fSup_add_less)
      apply (metis fSup_finite_join fSup_singleton insert_is_Un sup.assoc)
      done
    
  end  
  
  context complete_lattice begin
  
    lemma Sup_add_less: "\<lbrakk>x \<in> X; \<forall>y\<in>X. Sup {y, x} = x\<rbrakk> \<Longrightarrow> Sup X = x"
      apply (auto simp: local.Sup_union_distrib)
      apply (simp flip: le_iff_sup)
      by (simp add: local.Sup_least local.Sup_upper order.antisym)
      
    lemma Sup_add_less': "\<lbrakk>x \<in> X; \<forall>y\<in>X. sup y x = x\<rbrakk> \<Longrightarrow> Sup X = x"
      using Sup_add_less by simp
  
    lemma "complete_lattice_from_Sup Sup"
      apply unfold_locales
      apply (auto simp: local.Sup_union_distrib)
      apply (simp flip: le_iff_sup)
      by (simp add: local.Sup_least local.Sup_upper order.antisym)


    lemma "sup a b = Sup {a,b}" by simp
    lemma "less_eq a b \<longleftrightarrow> sup a b = b" by (rule le_iff_sup)
    lemma "Inf X = Sup {b. \<forall>a\<in>X. less_eq b a}" by (rule Inf_eq_Sup)   
    lemma "inf a b = Inf {a,b}" by simp
    lemma "less a b \<longleftrightarrow> less_eq a b \<and> a\<noteq>b" by auto
    lemma "top = Sup UNIV" by simp
    lemma "bot = Sup {}" by simp
  end
  

  subsection \<open>Interference Type\<close>
  text \<open>This type formalizes potential interference created by a thread. 
    On parallel combinations, the interferences must be disjoint.
  \<close>
  class interference = comm_monoid_add +
    fixes nointer :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
    assumes nointer_sym[sym]: "nointer x y \<Longrightarrow> nointer y x"
    assumes nointer_zero_left[simp, intro!]: "nointer 0 x"
    assumes nointer_add_left[simp]: "nointer (a+b) c \<longleftrightarrow> nointer a c \<and> nointer b c"
  begin
    lemma nointer_zero_right[simp, intro!]: "nointer x 0"
      apply (rule nointer_sym)
      by simp
      
    lemma nointer_add_right[simp]: "nointer a (b+c) \<longleftrightarrow> nointer a b \<and> nointer a c"  
      using nointer_add_left nointer_sym by blast
  
  end
  
  text \<open>We instantiate the typeclass right away, 
    to avoid typeclass dependencies in code equations later on\<close>
  instantiation unit :: interference
  begin
    definition [simp]: "nointer_unit = (top::unit \<Rightarrow> unit \<Rightarrow> bool)"
    definition [simp]: "0 \<equiv> ()"
    definition [simp]: "a+b \<equiv> ()"
    
    instance
      apply standard
      apply (simp_all)
      done
  
  end

  subsection \<open>Monad Type Definitions\<close>
  datatype ('a,'i::interference) mres = ismfail: mfail | mspec (the_mspec: "'a \<Rightarrow> 'i \<Rightarrow> bool")
  
  text \<open>Defined here such that automation which is used early can depend on state monad already\<close>
  datatype ('a,'i::interference,'s) M = M (run: "'s \<Rightarrow> ('a\<times>'s,'i) mres")

  subsection \<open>Nondetermistic Result Monad\<close>
  subsubsection \<open>Pointwise Reasoning\<close>
  
  named_theorems mpw_simp \<open>Pointwise reasoning simplification rules\<close>
  named_theorems mpw_init \<open>Pointwise reasoning initialization rules\<close>
  
  fun ismres where
    "ismres mfail _ _ = True"
  | "ismres (mspec R) r i = R r i"
  
  lemmas [mpw_simp] = ismres.simps
  lemma [mpw_simp]: "ismfail mfail" by simp
  
  subsubsection \<open>Automation\<close>  
  
  
  
  ML \<open>
  
    structure Monad_Util = struct
      (* TODO: The try_xxx functions are copied from RF2_Util *)
      fun try' hnd f x = SOME (f x) handle exn => if Exn.is_interrupt exn then Exn.reraise exn else hnd x exn;
    
      fun try_warn msg = try' (fn _ => fn exn => let
        val msg' = Runtime.exn_message_list exn |> cat_lines 
        val _ = warning (msg ^ "\n\n" ^ msg')
        (*
        val msgs = [@{make_string} exn |> Pretty.str] (*Runtime.exn_message_list exn |> map Pretty.str*)
        val pt = Pretty.fbreaks (Pretty.str msg :: msgs) |> Pretty.block
        val _ = warning (Pretty.string_of pt)
        *)
      in NONE end)
    
      fun try_repeat f x = case try f x of SOME x => try_repeat f x | NONE => x

      fun dest_def_thm thm = case Thm.prop_of thm of
        (@{const Trueprop}$(Const (@{const_name "HOL.eq"},_)$lhs$rhs)) => (lhs,rhs)
      | (Const (@{const_name Pure.eq},_)$lhs$rhs) => (lhs,rhs)  
      | _ => raise THM("dest_def_thm",1,[thm])
      
            
      val rpt_fun_cong = try_repeat (fn thm => @{thm fun_cong} OF [thm])
      fun def_to_HOL_eq thm = case Thm.concl_of thm of
        @{mpat \<open>_\<equiv>_\<close>} => @{thm HOL.meta_eq_to_obj_eq} OF [thm]
      | @{mpat \<open>Trueprop (_=_)\<close>} => thm
      | _ => raise THM ("def_to_HOL_eq",1,[thm])
      
      val norm_def_thm = def_to_HOL_eq #> rpt_fun_cong
      
      
      fun is_MT (Type (@{type_name M},_)) = true
        | is_MT _ = false
        
      val is_Mt = fastype_of #> is_MT
      
      val is_MfT = body_type #> is_MT
      val is_Mft = fastype_of #> is_MfT
      
      fun head_of_def_lhs (f$Var _) = head_of_def_lhs f
        | head_of_def_lhs t = t
      
      val head_of_def_thm = dest_def_thm #> fst #> head_of_def_lhs
      
      fun discharge_monos ctxt thm = let
        fun aux ctxt thm = let
          val prems = Thm.prems_of thm
  
          fun prove_simple tac t ctxt = Goal.prove ctxt [] [] t (fn {context=ctxt, ...} => ALLGOALS (tac ctxt))
  
  
          (*val mono_tac = Subgoal.FOCUS (fn {context=ctxt,...} => CHANGED (ALLGOALS (Partial_Function.mono_tac ctxt)))*)
          fun mono_tac ctxt = CHANGED o (Partial_Function.mono_tac ctxt)
  
          fun cinst (t as @{mpat "\<And>_. monotone (fun_ord _) _ _"}) = the_default asm_rl (try (prove_simple mono_tac t) ctxt)
            | cinst _ = asm_rl
  
          val insts = map cinst prems
  
          val thm = thm OF insts
        in
          thm
        end
      in
        (* Avoid surprises with schematic variables being instantiated *)
        singleton (Variable.trade (map o aux) ctxt) thm
      end
      
      val mono_tac = Subgoal.FOCUS_PREMS (fn {context=ctxt,...} => CHANGED (ALLGOALS (Partial_Function.mono_tac ctxt THEN_ALL_NEW TRY o assume_tac ctxt)))      
      
    end
  \<close>
  
  method_setup pf_mono_prover = \<open>Scan.succeed (SIMPLE_METHOD' o Monad_Util.mono_tac)\<close>

  attribute_setup discharge_monos
    = \<open>Scan.succeed (Thm.rule_attribute [] (Monad_Util.discharge_monos o Context.proof_of))\<close>
    \<open>Try to discharge monotonicity premises\<close>
  
  
  
  context begin  
    qualified definition "PWPROPS m q1 q2 \<equiv> 
        (ismfail m = q1) 
      \<and> (\<forall>r. ismres m r = q2 r)
      "
    
    qualified lemma PWPROPSI:
      assumes "EQ_TAG (ismfail m) (F)"
      assumes "\<And>r i. EQ_TAG (ismres m r i) (R r i)"
      shows "PWPROPS m F R"
      using assms unfolding PWPROPS_def EQ_TAG_def by auto
    
    qualified lemma PWPROPS_trivs: 
      "\<And>m F R. PWPROPS m F R \<Longrightarrow> PWPROPS m F R"
      "\<And>m s F R. PWPROPS (run m s) F R \<Longrightarrow> PWPROPS (run m s) F R"
      .
    
    qualified lemma PWPROPSD:
      assumes "PWPROPS m F R"
      shows "ismfail m = F" 
        and "ismres m r i = R r i" 
      using assms unfolding PWPROPS_def by (auto)+
      
    ML \<open>
    
      signature RF2_POINTWISE = sig
        (* Make mpw_simp theorem from definition theorem *)
        val mk_pw_props: Proof.context -> thm -> thm list 
        
        (* Declare mpw_simp from definition theorem *)
        val declare_pw_props: thm -> Context.generic -> Context.generic 
        
        (* Attribute to declare mpw_simp from definition theorem *)
        val mkpw_attrib: attribute context_parser
      end  
      
      structure RF2_Pointwise = struct
      
        fun mk_pw_props ctxt def_thm = let
          open Refine_Util Monad_Util
          val ctxt0 = ctxt
          val (def_thm,ctxt) = 
            yield_singleton (apfst snd oo Variable.import true) def_thm ctxt
            |>> Monad_Util.norm_def_thm
          
          val (t,_) = dest_def_thm def_thm |>> Thm.cterm_of ctxt
        
          fun inst trthm = trthm
            |> Drule.infer_instantiate' ctxt [SOME t]
            |> Thm.concl_of
        
          fun inst_first [] = raise CTERM("Definition must define constant of type mres or M",[t])
            | inst_first (x::xs) = case try inst x of NONE => inst_first xs | SOME r => r
              
          
          val goal = inst_first @{thms PWPROPS_trivs}
          
          val mpw_simp = Named_Theorems.get ctxt (@{named_theorems mpw_simp})
            
          fun tac ctxt = 
            SELECT_GOAL (unfold_tac ctxt [def_thm])
            THEN' resolve_tac ctxt @{thms PWPROPSI}  
            THEN_ALL_NEW (
              simp_tac (Splitter.del_split @{thm if_split} (ctxt addsimps mpw_simp)) 
              THEN' TRY o resolve_tac ctxt @{thms EQI}
            )
            
          val thm = Goal.prove ctxt [] [] goal (fn {context = ctxt, ...} => ALLGOALS (tac ctxt))
            |> singleton (Variable.export ctxt ctxt0)
           
          val thms = map (fn thm2 => thm2 OF [thm]) @{thms PWPROPSD}  
        in 
          thms
        end
      
        fun is_reduced_pointwise_prop (_:Proof.context) thm = let
          val (_,rhs) = Thm.concl_of thm |> HOLogic.dest_Trueprop |> HOLogic.dest_eq
          
          fun is_prim (Var _) = true
            | is_prim (Bound _) = true (* TODO: Should we warn on higher-order? *)
            | is_prim @{mpat \<open>run ?m _\<close>} = is_prim m
            | is_prim t = is_Var (head_of t)
          
          fun is_red @{mpat "ismfail ?m"} = is_prim m
            | is_red @{mpat "ismres ?m"} = is_prim m
            | is_red _ = true
        
        in
          not (exists_subterm (not o is_red) rhs)
        end
        
        
        fun declare_pw_props def_thm context = let
          val ctxt = Context.proof_of context
          val thms = Monad_Util.try_warn "Failed to generate pointwise definition" (mk_pw_props ctxt) def_thm
            |> the_default []
          
          fun warn_nonred thm = if is_reduced_pointwise_prop ctxt thm then true
            else (
              Pretty.block [Pretty.str "Ignoring generated non-reduced pointwise theorem", Pretty.brk 1, Thm.pretty_thm ctxt thm]
              |> Pretty.string_of |> warning; 
              false)
          
          val thms = filter warn_nonred thms
          
          (*val _ = map (Thm.pretty_thm ctxt) thms |> Pretty.big_list "Pointwise Reasoning Theorems (mpw_simp)" |> Pretty.string_of |> tracing*)
          val context = fold (Named_Theorems.add_thm @{named_theorems mpw_simp}) thms context
        in
          context
        end
      
        
        val mkpw_attrib : attribute context_parser = Scan.succeed (Thm.declaration_attribute declare_pw_props) 
        

        (*        
        fun mpw_tac ctxt = let
          val ctxt = ctxt addsimps Named_Theorems.get ctxt @{named_theorems mpw_simp}
        in
          asm_full_simp_tac ctxt
          THEN_ALL_NEW TRY o (blast_tac ctxt)
        end
        *)
        
      end  
    \<close>  
    
    attribute_setup rf2_mkpw = \<open>RF2_Pointwise.mkpw_attrib\<close>
                                                                                                             
    method_setup mpw_simp_only = \<open>Scan.succeed (fn ctxt => SIMPLE_METHOD' (CHANGED o asm_full_simp_tac (
      put_simpset HOL_ss ctxt 
      addsimprocs [@{simproc unit_eq}]
      addsimps Named_Theorems.get ctxt @{named_theorems mpw_simp}
    )))\<close>


    lemma fold_mfail:
      "a=mfail \<longleftrightarrow> ismfail a"
      "mfail=a \<longleftrightarrow> ismfail a"
      by (cases a; simp)+
    
    method mpw_propagate_fail = changed \<open>(unfold ismfail_def; simp; (unfold fold_mfail)?)\<close>
        
    
    
    method mpws methods m = ((unfold mpw_init)?; repeat_new \<open>(mpw_simp_only|mpw_propagate_fail)\<close>; m)
    method mpw = mpws blast
    
    method mpwt = mpws \<open>blast?\<close>
    
    method mpws' methods m = changed \<open>((simp add: mpw_init)?; (repeat_new \<open>simp add: mpw_simp | mpw_propagate_fail\<close>)?; m)\<close>
    method mpw' = mpws' blast
    method mpw't = mpws' \<open>blast?\<close>
    
    method mpwsa methods m = changed \<open>
      ((auto split: if_splits simp: mpw_init)?; 
      (repeat_new \<open>auto split: if_splits simp: mpw_simp | mpw_propagate_fail\<close>)?;
       m)\<close>
    method mpwa = mpwsa blast
    method mpwat = mpwsa \<open>blast?\<close>
      
  end

  subsubsection \<open>Init-Rules for Equality\<close>
  
  lemma mres_eq_iff[mpw_init]: "mr = mr' \<longleftrightarrow> (\<forall>r i. ismfail mr = ismfail mr' \<and> ismres mr r i = ismres mr' r i)"
    apply (cases mr; cases mr') by (auto del: ext intro!: ext)
  
  lemmas [simp del] = ismres.simps mres.disc mres.collapse
  lemmas [mpw_simp] = ismres.simps mres.disc 
  
  subsubsection \<open>Standard Lemmas\<close>
  
  (*lemmas [mpw_simp] = ismfail_def *) 
    
  (*lemma [mpw_simp]: "NO_MATCH mfail m \<Longrightarrow> mfail = m \<longleftrightarrow> m = mfail" by auto *)
    
  
  (* Some safe rules for standard theorem sets *)      
  lemma is_fail_imp_nomres[intro,simp]:
    "ismfail m \<Longrightarrow> ismres m r i"    
    by (cases m; auto simp: mpw_simp; fail)+

  (* TODO: Automation for splitting in mpw! *)
  lemmas std_splits = if_split prod.split option.split
    
  lemmas mpw_m_std_splits[mpw_simp] =  
    std_splits[where P="ismfail"]  
    std_splits[where P="\<lambda>m. ismres m _ _"]
    
  lemmas mpw_M_std_splits[mpw_simp] =  
    std_splits[where P="\<lambda>m. ismfail (run m _)"]  
    std_splits[where P="\<lambda>m. ismres (run m _) _ _"]
  
  
  subsubsection \<open>Instantiation as Complete Lattice\<close>

  instantiation mres :: (type,interference) complete_lattice
  begin
    definition "Sup X = (if mfail\<in>X then mfail else mspec (Sup {R . mspec R \<in> X}))"

    lemma mpw_Sup[mpw_simp]:
      "ismfail (Sup X) \<longleftrightarrow> mfail \<in> X"
      "ismres (Sup X) r i \<longleftrightarrow> (\<exists>x\<in>X. ismres x r i)"
      unfolding Sup_mres_def 
      apply mpw 
      apply (mpws' \<open>simp?\<close>)
      by (metis ismres.elims(2) ismres.simps(1) ismres.simps(2))
      
    lemma Sup_mres_alt: "Sup X = (if mfail \<in> X then mfail else mspec (Sup (the_mspec`X)))"
      unfolding Sup_mres_def 
      apply (auto simp: fun_eq_iff)
      apply force
      by (metis mres.collapse)
      
        
    interpretation complete_lattice_from_Sup "Sup::_ \<Rightarrow> ('a, 'b) mres"
      apply standard
      unfolding Sup_mres_def
      subgoal for x by (cases x) auto
      subgoal by auto
      subgoal for x X 
        apply (cases x) 
        apply (auto split: if_splits del: ext intro!: ext)
        by (smt (z3) Sup2_I mem_Collect_eq)
        
      done
    
    definition "Inf = fInf"
    definition [rf2_mkpw]: "bot = mspec (\<lambda>_ _. False)"
    
    lemma bot_mres_alt: "bot = fbot"
      unfolding fbot_def bot_mres_def Sup_mres_def by auto
    
    definition "sup = fsup"
    
    lemma sup_mRES[simp]: "sup (mspec X) (mspec Y) = mspec (\<lambda>x i. X x i \<or> Y x i)"
      unfolding sup_mres_def fsup_def Sup_mres_def by auto
    
    definition [rf2_mkpw]: "top = mfail"
    lemma top_mres_alt: "top = ftop"
      unfolding top_mres_def ftop_def Sup_mres_def by simp
    
    definition "inf = finf"
    
    lemma set2_aux: "{Ra. Ra = R \<or> Ra = X} = {R,X}" for R X :: "'a \<Rightarrow> bool" by auto
    
    
    lemma inf_mRES[simp]: "inf (mspec X) (mspec Y) = mspec (\<lambda>x i. X x i \<and> Y x i)"
    proof -
      show ?thesis
        unfolding inf_mres_def fInf_def finf_def fless_eq_def fsup_def Sup_mres_def 
        apply (simp split: if_splits add: set2_aux)
        apply (intro ext)
        apply safe
        subgoal by (metis (mono_tags, lifting) Sup2_I mem_Collect_eq)
        subgoal by (metis (mono_tags, lifting) Sup2_I mem_Collect_eq)
        subgoal by (auto intro: exI[where x="inf X Y"])
        done
    qed
    
    definition "less_eq = fless_eq"
    
    lemma less_eq_mRES_simps[simp]:
      "mspec X \<le> mspec Y \<longleftrightarrow> (\<forall>x i. X x i \<longrightarrow> Y x i)"
      unfolding less_eq_mres_def fless_eq_def fsup_def Sup_mres_def
      apply (auto simp: set2_aux)
      by (metis (mono_tags, lifting) Sup2_I mem_Collect_eq)
    
    definition "less = fless"
    
    instance
      apply (rule class.Complete_Lattices.complete_lattice.of_class.intro)
      unfolding Inf_mres_def bot_mres_alt sup_mres_def top_mres_alt inf_mres_def less_eq_mres_def less_mres_def
      by (rule is_complete_lattice)

    lemma mpw_sup_mres[mpw_simp]: 
      "ismfail (sup a b) \<longleftrightarrow> ismfail a \<or> ismfail b"  
      "ismres (sup a b) r i \<longleftrightarrow> ismres a r i \<or> ismres b r i"  
      unfolding sup_mres_def fsup_def
      apply (simp_all only: mpw_simp)
      apply (cases a; cases b; mpw)
      apply (cases a; cases b; mpwa)
      done

    lemma mpw_less_eq_mres[mpw_init]:
      "a\<le>b \<longleftrightarrow> ismfail b \<or> \<not>ismfail a \<and> (\<forall>x i. ismres a x i \<longrightarrow> ismres b x i)"  
      unfolding le_iff_sup
      by mpwa
      
      
    lemma all_in_set_eq_conv: "(\<forall>x\<in>X. x=c) \<longleftrightarrow> X={} \<or> X={c}"
      by auto  
      
    lemma ismres_eta_simps[mpw_simp]: 
      "ismres mfail = (\<lambda>_ _. True)" 
      "ismres (mspec P) = P"
      by (intro ext; mpw)+  
      
    
    lemma mpw_Inf_mres[mpw_simp]: 
      "ismfail (Inf X) \<longleftrightarrow> (\<forall>x\<in>X. ismfail x)"
      "ismres (Inf X) r i \<longleftrightarrow> (\<forall>x\<in>X. ismres x r i)"
      unfolding Inf_eq_Sup less_eq_mres_def fless_eq_def fsup_def
      supply [simp] = all_in_set_eq_conv
      apply mpwat
      apply safe 
      apply mpwat
      apply mpw_simp_only
      apply (rule bexI[where x="if X={} then mfail else mspec (\<lambda>r' i'. r'=r \<and> i'=i)"])
      apply mpw
      apply mpw'
      done

    lemma inf_mres_as_Inf: "inf a b = Inf {a,b::(_,_) mres}" by auto 
      
    lemma mpw_inf_mres[mpw_simp]:
      "ismfail (inf a b) \<longleftrightarrow> ismfail a \<and> ismfail b"  
      "ismres (inf a b) r i \<longleftrightarrow> ismres a r i \<and> ismres b r i"  
      unfolding inf_mres_as_Inf
      apply mpw 
      apply mpw 
      done
      
    lemma mpw_less_mres[mpw_init]: 
      "a<b \<longleftrightarrow> (\<forall>x i. ismres a x i \<longrightarrow> ismres b x i) \<and> \<not>ismfail a \<and> (\<not>ismfail b \<longrightarrow> (\<exists>x i. \<not>ismres a x i \<and> ismres b x i))"  
      unfolding less_le
      by mpw'
            
  end    
  
  
  subsubsection \<open>Monad Operations\<close>
  
  definition [rf2_mkpw]: "mreturn x \<equiv> mspec (\<lambda>r i. r = x \<and> i=0)"
  
  fun minter where
    "minter mfail i = mfail"
  | "minter (mspec P) i = mspec (\<lambda>x i'. \<exists>i''. P x i'' \<and> i'=i+i'')"  
  
  lemma mpw_minter[mpw_simp]:
    "ismfail (minter m i) = ismfail m"
    "ismres (minter m i) x i' \<longleftrightarrow> ismfail m \<or> (\<exists>i''. ismres m x i'' \<and> i'=i+i'')"
    by (cases m; simp; mpw)+
    
    
  fun mbind where
    "mbind mfail _ = mfail"
  | "mbind (mspec P) f = Sup {minter (f x) i | x i. P x i}"  

  
  
  lemma mpw_mbind[mpw_simp]:
    "ismfail (mbind m f) \<longleftrightarrow> ismfail m \<or> (\<exists>x i. ismres m x i \<and> ismfail (f x))"
    "ismres (mbind m f) r i \<longleftrightarrow> ismfail m \<or> (\<exists>x i'. ismres m x i' \<and> (ismres (minter (f x) i') r i))"
    (*"ismres (mbind m f) r i \<longleftrightarrow> ismfail m \<or> (\<exists>x i'. ismres m x i' \<and> (ismfail (f x) \<or> (\<exists>i''. ismres (f x) r i'' \<and> i=i'+i'')))"*)
    apply (cases m; simp; mpw')
    apply (cases m; simp)
    apply mpw'
    subgoal premises for ms
    apply (mpws succeed)
    apply (auto simp: algebra_simps)
    apply blast
    by (smt (z3) mpw_minter(1) mpw_minter(2))
    done
  
  
  (*
  lemma mbind_mRES_eq[simp]: "mbind (mRES M) f = Sup (f`M)"
    unfolding mRES_def by (auto simp: Setcompr_eq_image)
  *)  
  
  lemma mbind_eq_fail_direct: "mfail = mbind m f \<longleftrightarrow> mbind m f = mfail" by auto 
    
  
  lemma m_monad_laws[simp]:
    "mbind (mreturn x) f = f x"
    "mbind m mreturn = m"
    "mbind (mbind m f) g = mbind m (\<lambda>x. mbind (f x) g)"
    apply mpw
    apply mpw
    apply (mpws \<open>fastforce simp: add.assoc\<close>)
    done

  subsubsection \<open>Recursion\<close>
  interpretation m: monad_rec mfail .

  lemma ex_set_aux_lemma: "{y. \<exists>x\<in>S. y = f x} = {f x | x. x\<in>S}" by blast 
  lemma ex_aux_lemma: "(\<forall>xa. (\<exists>s. xa = f s \<and> S s) \<longrightarrow> P xa) = (\<forall>s. S s \<longrightarrow> P (f s))" by auto
  
  lemma mpw_leq_iff[mpw_init]: "m.leq m m' \<longleftrightarrow> (ismfail m 
    \<or> (\<not>ismfail m' \<and> (\<forall>r' i'. ismres m r' i' = ismres m' r' i')))"
    unfolding flat_ord_def
    by (mpw)
    
  (* Note the similarity to refinement ordering! Actually, for chains, Sup should coincide! *)  
  lemma mpw_flat_lub[mpw_simp]:
    assumes "Complete_Partial_Order.chain m.leq C" 
    shows "ismfail (m.lub C) = (\<forall>x\<in>C. ismfail x)"  
          "ismres (m.lub C) r i = (\<forall>x\<in>C. ismres x r i)"  
    using assms    
    by (auto elim!: flat_chain_cases simp: mpw_simp)
  
  text \<open>For chains, we can switch between flat ordering and refinement ordering!\<close>  
  theorem PW_lub_compat: "Complete_Partial_Order.chain m.leq C \<Longrightarrow> m.lub C = Inf C"
    by mpw'
    
  theorem PW_fun_lub_compat: "Complete_Partial_Order.chain m.PF.le_fun C \<Longrightarrow> m.PF.lub_fun C x = Inf C x"  
    apply (simp only: fun_lub_def Inf_fun_def)
    apply (subst PW_lub_compat)
    subgoal by (auto simp: chain_def fun_ord_def)
    subgoal apply (rule arg_cong[where f=Inf]) by auto
    done
    
  (* TODO: General lemma *)  
  lemma fun_flat_chain_imp_flat: "Complete_Partial_Order.chain (fun_ord (flat_ord fail)) C \<Longrightarrow> Complete_Partial_Order.chain (flat_ord fail) {f x |f. f \<in> C}"  
    unfolding chain_def 
    apply (simp only: le_fun_def fun_ord_def) 
    by blast
    
  theorem fun_mres_lub_compat: "Complete_Partial_Order.chain (fun_ord (flat_ord mfail)) C \<Longrightarrow> fun_lub (flat_lub mfail) C = Inf C"  
    apply (rule ext)
    apply (rule mres_eq_iff[THEN iffD2])
    apply (simp only: fun_lub_def Inf_fun_def fun_eq_iff ex_set_aux_lemma mpw_simp fun_flat_chain_imp_flat)
    by blast

  subsubsection \<open>Continuity and Admissibilty\<close>  
        
  lemma leq_implies_refine: "m.leq m m' \<Longrightarrow> m' \<le> m"  
    by mpw'
    
  (*  
  lemma mcont_refine_id: "mcont m.lub m.leq Inf (\<ge>) (\<lambda>x. x)"
    apply (intro mcontI monotoneI contI leq_implies_refine)
    apply simp
    apply simp 
    apply (simp add: PW_lub_compat)
    done
  *)    

    
  subsection \<open>State Monad\<close>  
      

  lemma run_eq_iff[simp]: "run a = run b \<longleftrightarrow> a=b"
    by (cases a; cases b; simp)
  
  
  subsubsection \<open>Complete Lattice Structure\<close>
  instantiation M :: (type,interference,type) complete_lattice
  begin
    
    definition "Sup X \<equiv> M (Sup (run`X))"
    interpretation complete_lattice_from_Sup "Sup :: _ \<Rightarrow> (_,_,_) M"
      apply standard
      unfolding Sup_M_def
      subgoal by auto
      subgoal by (auto simp: SUP_union)
      subgoal for x X 
        apply (cases x; simp add: fun_eq_iff)
        apply (force simp del: Sup_insert intro: Sup_add_less')
        done
      done
  
    definition "Inf X = M (Inf (run`X))"
    definition "bot = M (\<lambda>_. bot)"
    definition "sup a b = M (sup (run a) (run b))"
    definition "top = M (\<lambda>_. mfail)"
    definition "inf a b = M (inf (run a) (run b))"
    definition "less_eq a b \<longleftrightarrow> run a \<le> run b"
    definition "less = fless"
    
    lemma top_M_alt: "top = ftop"
      unfolding top_M_def ftop_def apply (rule sym) unfolding Sup_M_def apply auto
      by (metis M.sel Sup_UNIV UNIV_eq_I rangeI top_apply top_mres_def)
    
    lemma bot_M_alt: "bot = fbot"
      unfolding bot_M_def fbot_def Sup_M_def by auto
    
    lemma less_eq_M_alt: "less_eq a b \<longleftrightarrow> fless_eq a b"
      unfolding less_eq_M_def fless_eq_def fsup_def Sup_M_def
      by (smt M.sel Sup_M_def SUP_insert fSup_singleton le_iff_sup)

    thm less_eq_M_alt[symmetric, unfolded less_eq_M_def]  
      
    lemma Inf_M_alt: "Inf X = fInf X"
      unfolding Inf_M_def fInf_def less_eq_M_alt[symmetric, unfolded less_eq_M_def]
      unfolding Sup_M_def Inf_eq_Sup
      apply (auto intro!: arg_cong[where f=Sup])
      by (smt M.sel image_iff mem_Collect_eq)

    lemma inf_M_alt: "inf = finf"
      unfolding finf_def Inf_M_alt[symmetric] Inf_M_def inf_M_def
      by simp
    
    lemma sup_M_alt: "sup = fsup"
      unfolding fsup_def 
      unfolding Sup_M_def sup_M_def
      by simp
      
          
    instance
      apply (rule class.Complete_Lattices.complete_lattice.of_class.intro)
      unfolding Inf_M_alt bot_M_alt sup_M_alt top_M_alt inf_M_alt less_eq_M_alt less_M_def
      by (rule is_complete_lattice)
      
      
    (* TODO: show how ordering is derived from underlying ordering! *)  
  end
  
  abbreviation "fail \<equiv> top::(_,_,_) M"
  
  lemma mpw_fail[mpw_simp]: 
    "ismfail (run fail s)"
    "ismres (run fail s) x i"
    unfolding top_M_def
    by mpw'+
  
    
  definition srel where [rf2_mkpw]: "srel R \<equiv> M (\<lambda>s. mspec (\<lambda>(x,s'). R s (x,s')))"
    
  
  definition spec where [rf2_mkpw]: "spec P Q \<equiv> M (\<lambda>s. if P s then mspec (\<lambda>(x,s). Q x s) else mfail)"

    
  lemma run_Inf[mpw_simp]: "run (Inf F) s = Inf ({run f s | f. f\<in>F})"
    unfolding Inf_M_def
    apply simp 
    apply (fo_rule arg_cong)
    by blast
    
  definition spec' where "spec' P Q \<equiv> Inf {spec (P xs) (Q xs) | xs. True}"
  
  lemma spec'_mfail[mpw_simp]:
    "ismfail (run (spec' P Q) s) = (\<forall>xs. ismfail (run (spec (P xs) (Q xs)) s))"
    unfolding spec'_def
    apply (mpws succeed)
    apply (simp flip: ex_simps all_simps)
    apply mpwt
    done
    
  lemma spec'_ismres[mpw_simp]:
    "ismres (run (spec' P Q) s) r i = (\<forall>xs. ismres (run (spec (P xs) (Q xs)) s) r i)"  
    unfolding spec'_def
    apply mpwt
    apply (simp flip: ex_simps all_simps)
    apply mpwt
    done
    
  
  subsubsection \<open>Monad Structure\<close>

  definition [rf2_mkpw]: "return x \<equiv> M (\<lambda>s. mreturn (x,s))"
  definition [rf2_mkpw]: "bind m f \<equiv> M (\<lambda>s. mbind (run m s) (\<lambda>(x,s). run (f x) s) )"

  definition [rf2_mkpw]: "get \<equiv> M (\<lambda>s. mreturn (s,s))"
  definition [rf2_mkpw]: "put s \<equiv> M (\<lambda>_. mreturn ((),s))"
  
  (*lemma run_return[mpw_simp]: "run (return x) s = mreturn (x, s)"
    by (simp add: return_def)

  lemma run_bind[mpw_simp]: "run (bind m f) s = mbind (run m s) (\<lambda>(x, y). run (f x) y)"
    by (simp add: bind_def)
      
  lemma run_get[mpw_simp]: "run get s = mreturn (s,s)"
    by (simp add: get_def)
    
  lemma run_put[mpw_simp]: "run (put s) s' = mreturn ((),s)"  
    by (simp add: put_def)
  *)  
    
  lemma monad_laws[simp]:
    "bind (return x) f = f x"
    "bind m return = m"
    "bind (bind m f) g = bind m (\<lambda>x. bind (f x) g)"
    unfolding bind_def return_def
    by (auto simp: fun_eq_iff case_prod_unfold)

    
  subsubsection \<open>Syntax\<close>  
  abbreviation (do_notation) bind_doN where "bind_doN \<equiv> bind"
  abbreviation (do_notation) then_doN where "then_doN m f \<equiv> bind_doN m (\<lambda>_. f)"

  notation bind_doN (infixr "\<bind>" 54)

  notation then_doN (infixr "\<then>" 54)

  nonterminal doN_binds and doN_bind
  syntax
    "_doN_block" :: "doN_binds \<Rightarrow> 'a" ("doN {//(2  _)//}" [12] 62)
    "_doN_bind"  :: "[pttrn, 'a] \<Rightarrow> doN_bind" ("(2_ \<leftarrow>/ _)" 13)
    "_doN_let" :: "[pttrn, 'a] \<Rightarrow> doN_bind" ("(2let _ =/ _)" [1000, 13] 13)
    "_doN_then" :: "'a \<Rightarrow> doN_bind" ("_" [14] 13)
    "_doN_final" :: "'a \<Rightarrow> doN_binds" ("_")
    "_doN_cons" :: "[doN_bind, doN_binds] \<Rightarrow> doN_binds" ("_;//_" [13, 12] 12)
    (*"_thenM" :: "['a, 'b] \<Rightarrow> 'c" (infixr "\<then>" 54)*)

  syntax (ASCII)
    "_doN_bind" :: "[pttrn, 'a] \<Rightarrow> doN_bind" ("(2_ <-/ _)" 13)
    (*"_thenM" :: "['a, 'b] \<Rightarrow> 'c" (infixr ">>" 54)*)

  translations
    "_doN_block (_doN_cons (_doN_then t) (_doN_final e))"
      \<rightleftharpoons> "CONST then_doN t e"

    "_doN_block (_doN_cons (_doN_bind p t) (_doN_final e))"
      \<rightleftharpoons> "CONST bind_doN t (\<lambda>p. e)"

    "_doN_block (_doN_cons (_doN_let p t) bs)"
      \<rightleftharpoons> "let p = t in _doN_block bs"

    "_doN_block (_doN_cons b (_doN_cons c cs))"
      \<rightleftharpoons> "_doN_block (_doN_cons b (_doN_final (_doN_block (_doN_cons c cs))))"

    "_doN_cons (_doN_let p t) (_doN_final s)"
      \<rightleftharpoons> "_doN_final (let p = t in s)"

    "_doN_block (_doN_final e)" \<rightharpoonup> "e"
  

  subsubsection \<open>Recursion\<close>  
    
  interpretation state_monad_rec run M mfail top
    by unfold_locales (auto simp: top_M_def)
    
    
    
  subsection \<open>Monotonicity Automation\<close>  
  ML \<open>
    structure RF2_Make_Mono = struct
      open Monad_Util
      
      
      
      fun mk_leq (a,b) = @{mk_term "leq ?a ?b"}
        
      
      fun mk_mono_stmt ctxt t1 t2 = let
        val ctxt0 = ctxt
        val T1 = fastype_of t1
        val T2 = fastype_of t2
        val _ = T1 = T2 orelse raise TYPE("mk_mono_stmt: different types",[T1,T2],[t1,t2])
        val T = T1
        val _ = is_MfT T orelse raise TYPE("mk_mono_stmt: Not a monadic function",[T],[t1,t2])
  
        fun fixpar name T ctxt = 
          yield_singleton Variable.variant_fixes name ctxt
          |>> (fn n => Free (n,T)) 
        
        fun mk_param T ctxt = 
          if is_MfT T then let
              val (p1,ctxt) = fixpar "m" T ctxt
              val (p2,ctxt) = fixpar "mm" T ctxt
              val asm = mk_mono_stmt ctxt p1 p2
            in (((p1,p2),[p1,p2],[asm]),ctxt) end
          else let
              val (p,ctxt) = fixpar "x" T ctxt
            in (((p,p),[p],[]),ctxt) end
  
        val Ts = binder_types T
        val (pps,ctxt) = fold_map mk_param Ts ctxt
        
        val ps1 = map (fst o #1) pps
        val ps2 = map (snd o #1) pps
        val fvs = map #2 pps |> flat
        val asms = map #3 pps |> flat    
  
        (*                  
          Assemble statement
          !!fvs. [| asms |] ==> leq (t1$$ps1) (t2$$ps2) 
        *)
        val res = mk_leq (list_comb (t1,ps1), list_comb (t2,ps2))
          |> HOLogic.mk_Trueprop
        val res = Logic.list_implies (asms,res)  
        val res = fold Logic.all fvs res
        
        val res = singleton (Variable.export_terms ctxt ctxt0) res   
        
      in
        res
      end
  
      fun needs_mono_thm def_thm = let
        val T = head_of_def_thm def_thm |> fastype_of
        val Ts = binder_types T
      in
        is_MfT T andalso exists is_MfT Ts
      end 
       
      fun mk_mono_thm ctxt def_thm = let
        val ctxt0 = ctxt
        
        val ct = head_of_def_thm def_thm
        
        val (ct,ctxt) = yield_singleton (Variable.import_terms true) ct ctxt
        val _ = is_Mft ct orelse raise TERM("mk_mono_thm: Not a monadic function",[ct])
        
        
        val goal = mk_mono_stmt ctxt ct ct
        
        fun tac ctxt = 
          SELECT_GOAL (unfold_tac ctxt [def_thm])
          THEN' Monad_Util.mono_tac ctxt
        
        val res = Goal.prove ctxt [] [] goal (fn {context = ctxt, ...} => ALLGOALS (tac ctxt))
        
        val res = singleton (Variable.export ctxt ctxt0) res
        
      in
        res
      end
      
      fun declare_mono_thm def_thm context =
        if needs_mono_thm def_thm then let
          val _ = error "temporarily disabled, state mono thm manually!"
          val ctxt = Context.proof_of context
          val thm = mk_mono_thm ctxt def_thm
          val context = Named_Theorems.add_thm @{named_theorems partial_function_mono} thm context
        in context end
      else (
        Pretty.block [Pretty.str "No mono thm required for definition", Pretty.brk 1, Thm.pretty_thm (Context.proof_of context) def_thm] 
          |> Pretty.string_of |> warning;
        context
      )
      
      val mkmono_attrib : attribute context_parser = Scan.succeed (Thm.declaration_attribute declare_mono_thm) 
    end
  \<close>    
  attribute_setup rf2_mkmono = RF2_Make_Mono.mkmono_attrib \<open>Declare monotonicity rule from definition\<close>
      
  
  thm partial_function_mono
  
  lemmas [mpw_init] = 
    monotone_def[of _ m.leq]
    monotone_def[of _ leq]
  
  lemma leq_eq_mleq[mpw_init]: "leq m m' \<longleftrightarrow> (\<forall>s. m.leq (run m s) (run m' s))"
    unfolding img_ord_def fun_ord_def ..
  
  
  lemma mbind_mono[partial_function_mono]: "monotone ord m.leq m \<Longrightarrow> (\<And>x. monotone ord m.leq (f x)) \<Longrightarrow> monotone ord m.leq (\<lambda>D. mbind (m D) (\<lambda>x. f x D))"
    by mpw  
    
  lemma bind_mono[partial_function_mono]: "monotone ord leq m \<Longrightarrow> (\<And>x. monotone ord leq (f x)) \<Longrightarrow> monotone ord leq (\<lambda>D. (m D) \<bind> (\<lambda>x. f x D))"
    by mpw
  
  subsection \<open>Weakest (Liberal) Precondition\<close>

  text \<open>We define a universal and an existential version\<close>  
  definition "gwp m s F Q \<equiv> (if ismfail (run m s) then F else \<forall>x i s'. ismres (run m s) (x,s') i \<longrightarrow> Q x s')"
  definition "cgwp m s F Q \<equiv> (if ismfail (run m s) then F else \<exists>x i s'. ismres (run m s) (x,s') i \<and> Q x s')"

  lemma gwp_mpw_init[mpw_init]: "gwp m s F Q \<longleftrightarrow> (ismfail (run m s) \<longrightarrow> F) \<and> (\<not>ismfail (run m s) \<longrightarrow> (\<forall>x s' i. ismres (run m s) (x,s') i \<longrightarrow> Q x s'))"
    unfolding gwp_def by auto 

  lemma cgwp_mpw_init[mpw_init]: "cgwp m s F Q \<longleftrightarrow> (ismfail (run m s) \<longrightarrow> F) \<and> (\<not>ismfail (run m s) \<longrightarrow> (\<exists>x s' i. ismres (run m s) (x,s') i \<and> Q x s'))"
    unfolding cgwp_def by auto 
    
      
  text \<open>Alternative representation, as basic as possible\<close>
  lemma gwp_alt: "gwp m s F Q = (case run m s of mfail \<Rightarrow> F | mspec R \<Rightarrow> (\<forall>x s' i. R (x,s') i \<longrightarrow> Q x s'))"
    apply (simp split: mres.splits)
    apply mpw'
    done

  lemma cgwp_alt: "cgwp m s F Q = (case run m s of mfail \<Rightarrow> F | mspec R \<Rightarrow> (\<exists>x s' i. R (x,s') i \<and> Q x s'))"
    apply (simp split: mres.splits)
    apply mpw'
    done
    
  
  lemma gwp_mono: "F\<le>F' \<Longrightarrow> Q\<le>Q' \<Longrightarrow> gwp m s F Q \<le> gwp m s F' Q'"
    by mpw'
    
  lemma cgwp_mono: "F\<le>F' \<Longrightarrow> Q\<le>Q' \<Longrightarrow> cgwp m s F Q \<le> cgwp m s F' Q'"
    by mpw'
    
  lemma gwp_combine: "gwp m s F Q \<Longrightarrow> gwp m s F' Q' \<Longrightarrow> gwp m s (inf F F') (inf Q Q')"  
    by mpw'
    
  lemma cgwp_combine: "cgwp m s (inf F F') (inf Q Q') \<Longrightarrow> cgwp m s F Q \<and> cgwp m s F' Q'"  
    by mpw'
    
  lemma gwp_conseq:
    assumes "gwp m s F Q"
    assumes "F \<Longrightarrow> F'"
    assumes "\<And>x s. Q x s \<Longrightarrow> Q' x s"
    shows "gwp m s F' Q'"
    using assms by mpw
    
  lemma cgwp_conseq:
    assumes "cgwp m s F Q"
    assumes "F \<Longrightarrow> F'"
    assumes "\<And>x s. Q x s \<Longrightarrow> Q' x s"
    shows "cgwp m s F' Q'"
    using assms by mpwa
    
  lemma gwp_fail: "gwp fail s F Q = F" by mpw
  lemma gwp_return: "gwp (return x) s F Q = Q x s" by mpw
  lemma gwp_bind: "gwp m s F (\<lambda>x s. gwp (f x) s F Q) \<Longrightarrow> gwp (doN {x\<leftarrow>m; f x}) s F Q" 
    by (mpws \<open>fastforce split: if_splits\<close>)

  lemma wp_bind: "gwp m s bot (\<lambda>x s. gwp (f x) s bot Q) \<longleftrightarrow> gwp (doN {x\<leftarrow>m; f x}) s bot Q" 
    by (mpws \<open>fastforce split: if_splits\<close>)
      
  lemma gwp_rec:
    assumes M: "mono_body B"
    assumes W: "wf V"
    assumes I: "P arb t x s"
    assumes S: "\<And>D arb t x s. \<lbrakk> P arb t x s; \<And>t' x' s'. \<lbrakk> P arb t' x' s'; (t',t)\<in>V \<rbrakk> \<Longrightarrow> gwp (D x') s' (F arb x' s') (Q arb x' s')\<rbrakk> \<Longrightarrow> gwp (B D x) s (F arb x s) (Q arb x s)"
    shows "gwp (rec B x) s (F arb x s) (Q arb x s)"
  proof -
    have "\<forall>x s arb. P arb t x s \<longrightarrow> gwp (rec B x) s (F arb x s) (Q arb x s)"  
      apply (rule basic_rec_wf_rule[OF M W])
      using S by blast
    with I show ?thesis by blast  
  qed

  lemma gwp_get: "gwp get s F Q \<longleftrightarrow> Q s s" by mpw
  lemma gwp_put: "gwp (put s') s F Q \<longleftrightarrow> Q () s'" by mpw
  
  lemma cgwp_fail: "cgwp fail s F Q = F" by mpw
  lemma cgwp_return: "cgwp (return x) s F Q = Q x s" by mpw
  
  text \<open>Note, this direction is not usable in VCG. However, for \<open>F=top\<close>, we get the equivalence\<close>
  lemma cgwp_bind: "cgwp (doN {x\<leftarrow>m; f x}) s F Q \<Longrightarrow> cgwp m s F (\<lambda>x s. cgwp (f x) s F Q)" 
    by (mpws \<open>fastforce split: if_splits\<close>)

  lemma cwlp_bind: "cgwp (doN {x\<leftarrow>m; f x}) s top Q \<longleftrightarrow> cgwp m s top (\<lambda>x s. cgwp (f x) s top Q)" 
    by (mpws \<open>fastforce split: if_splits\<close>)
    
      
  lemma cgwp_rec:
    assumes M: "mono_body B"
    assumes W: "wf V"
    assumes I: "P arb t x s"
    assumes S: "\<And>D arb t x s. \<lbrakk> P arb t x s; \<And>t' x' s'. \<lbrakk> P arb t' x' s'; (t',t)\<in>V \<rbrakk> \<Longrightarrow> cgwp (D x') s' (F arb x' s') (Q arb x' s')\<rbrakk> \<Longrightarrow> cgwp (B D x) s (F arb x s) (Q arb x s)"
    shows "cgwp (rec B x) s (F arb x s) (Q arb x s)"
  proof -
    have "\<forall>x s arb. P arb t x s \<longrightarrow> cgwp (rec B x) s (F arb x s) (Q arb x s)"  
      apply (rule basic_rec_wf_rule[OF M W])
      using S by blast
    with I show ?thesis by blast  
  qed

  lemma cgwp_get: "cgwp get s F Q \<longleftrightarrow> Q s s" by mpw
  lemma cgwp_put: "cgwp (put s') s F Q \<longleftrightarrow> Q () s'" by mpw
  
  
  find_theorems run bind
  
  
(*  
  subsection \<open>Data Refinement\<close>
  
  definition [mpw_init]: "rel_mres R mi m \<equiv> ismfail m \<or> (\<not>ismfail mi \<and> (\<forall>xi i. ismres mi xi \<longrightarrow> (\<exists>x. ismres m x \<and> R xi x)))"
  
  fun conc_fun where 
    "conc_fun R mfail = mfail"
  | "conc_fun R (mspec P) = mspec (\<lambda>x. \<exists>y. R x y \<and> P y)"  
  
  lemma mpw_conc_fun[mpw_simp]:  
    "ismfail (conc_fun R m) \<longleftrightarrow> ismfail m"
    "ismres (conc_fun R m) x \<longleftrightarrow> ismfail m \<or> (\<exists> y. R x y \<and> ismres m y)"
    by (cases m; simp; mpw)+

  lemma "rel_mres R mi m \<longleftrightarrow> mi \<le> conc_fun R m"      
    by mpw'
    
  text \<open>Data refinement can also be used to express Hoare-style correctness\<close>  
  lemma le_mspec_rel_mres_conv: "m \<le> mspec Q \<longleftrightarrow> rel_mres (\<lambda>x _. Q x) m (mspec top)" by mpw
    
    
  subsubsection \<open>Structural Rules\<close>  
  
  text \<open>Refinement with additional post-condition. Used to structure refinement proofs.\<close>
  definition [mpw_init]: "mrefinec mi m Q \<equiv> rel_mres Q mi m"  
  
  text \<open>Refinement in state monad\<close>
  (*definition [mpw_init]: "refine R mi m si s \<equiv> rel_mres (\<lambda>(ri,si) (r,s). R ri r si s) (run mi si) (run m s)"*)
  definition [mpw_init]: "refinec mi m si s Q \<equiv> mrefinec (run mi si) (run m s) (\<lambda>(ri,si) (r,s). Q ri r si s)"

  
  text \<open>Refinement as weakest precondition\<close>  
  lemma refinec_as_wp: 
    "refinec mi m si s Q \<longleftrightarrow> gwp mi si (cgwp m s top bot) (\<lambda>ri si'. cgwp m s top (\<lambda>r s'. Q ri r si' s'))"
    by mpw't 
  
  (*
  lemma refine_mfail: "mrefinec mi mfail Q" by mpw
  lemma refine_mreturn:
    assumes "Q xi x"
    shows "mrefinec (mreturn xi) (mreturn x) Q"
    using assms by mpw
  
  lemma refine_mbind:
    assumes "mrefinec mi m (\<lambda>xi x. mrefinec (fi xi) (f x) Q)"
    shows "mrefinec (mbind mi fi) (mbind m f) Q"
    using assms by mpw
  
  lemma refine_mrec: 
    assumes M: "m.mono_body Fi"
    assumes P: "Ra xi x"
    assumes R: "\<And>Di D xi x. \<lbrakk> \<And>xi x. \<lbrakk>Ra xi x\<rbrakk> \<Longrightarrow> mrefinec (Di xi) (D x) Q; Ra xi x \<rbrakk> 
                \<Longrightarrow> mrefinec (Fi Di xi) (F D x) Q"
    shows "mrefinec (m.rec Fi xi) (m.rec F x) Q"
    using P
    apply (induction F x arbitrary: xi rule: m.pointwise_rec_induct)
    subgoal by mpw
    subgoal for D x xi
      apply (subst m.rec_unfold[OF M])
      by (rule R)
    done
  *)
    
  lemma refine_fail: "refinec m fail si s Q" by mpw
    
  lemma refine_return: "refinec (return xi) (return x) si s Q \<longleftrightarrow> Q xi x si s" by mpw
    
  lemma refine_bind:
    assumes "refinec mi m si s (\<lambda>xi x si s. refinec (fi xi) (f x) si s Q)"
    shows "refinec (doN {xi\<leftarrow>mi; fi xi}) (doN {x\<leftarrow>m; f x}) si s Q"
    using assms by mpw

  lemma refine_get:
    shows "refinec get get si s Q = Q si s si s"  
    by mpw
    
  lemma refine_put:
    shows "refinec (put si) (put s) si' s' Q = Q () () si s"
    by mpw
    
  lemma refine_rec:
    assumes M: "mono_body Fi"
    assumes P: "Ra xi si x s"
    assumes R: "\<And>Di D xi si x s. \<lbrakk> 
      \<And>xi si x s. \<lbrakk>Ra xi si x s\<rbrakk> \<Longrightarrow> refinec (Di xi) (D x) si s Q; 
      Ra xi si x s 
    \<rbrakk> \<Longrightarrow> refinec (Fi Di xi) (F D x) si s Q"
    shows "refinec (rec Fi xi) (rec F x) si s Q"
    using P unfolding refinec_def
    apply (induction F x s arbitrary: xi si rule: pointwise_rec_induct)
    subgoal by mpw
    subgoal for D x s xi si
      apply (subst rec_unfold[OF M])
      by (rule R[unfolded refinec_def])
    done

  lemma refine_if:
    assumes "bi=b"
    assumes "\<lbrakk> bi;b \<rbrakk> \<Longrightarrow> refinec ti t si s Q"
    assumes "\<lbrakk> \<not>bi;\<not>b \<rbrakk> \<Longrightarrow> refinec ei e si s Q"
    shows "refinec (if bi then ti else ei) (if b then t else e) si s Q"  
    using assms by auto
    
  lemma refine_let:  
    assumes "refinec (fi ai) (f a) si s Q"
    shows "refinec (let xi=ai in fi xi) (let x=a in f x) si s Q"
    using assms by auto
  
  (* TODO: Automation for case-splitting on refine *)  
  lemma refine_prod_case:
    assumes "\<And>ai bi a b. \<lbrakk> pi=(ai,bi); p=(a,b) \<rbrakk> 
      \<Longrightarrow> refinec (fi ai bi) (f a b) si s Q"
    shows "refinec (case pi of (ai,bi) \<Rightarrow> fi ai bi) (case p of (a,b) \<Rightarrow> f a b) si s Q"
    using assms by (auto split: prod.splits)
  
  
  text \<open>Asymmetric bind-refine rules\<close>  
    
  lemma refine_bind_left:
    assumes "\<not>ismfail (run mi si)"
    assumes "\<And>xi si'. ismres (run mi si) (xi,si') \<Longrightarrow> refinec (fi xi) c si' s Q"
    shows "refinec (doN {xi\<leftarrow>mi; fi xi}) c si s Q"
    using assms
    by mpw

  lemma refine_bind_right:
    assumes "\<not>ismfail (run m s) \<Longrightarrow> \<exists>x s'. ismres (run m s) (x,s') \<and> refinec ci (f x) si s' Q"
    shows "refinec ci (doN {x\<leftarrow>m; f x}) si s Q"
    using assms
    by mpw
  
*)    
    
(*    
  subsection \<open>Hoare-Style Rules\<close>
  
  lemma rel_mres_eq_is_le: "rel_mres (=) mi m \<longleftrightarrow> mi\<le>m" by mpw
    
  lemma "(\<exists>x. P x \<and> R xi x) \<Longrightarrow> rel_mres R (mreturn xi) (mspec P)" by mpw
  
  lemma spec_mreturn: "P x \<Longrightarrow> mreturn x \<le> mspec P" by mpw
  lemma spec_mbind: "m \<le> mspec (\<lambda>x. f x \<le> mspec P) \<Longrightarrow> mbind m (\<lambda>x. f x) \<le> mspec P" by mpw
  
  lemma spec_mrec:
    assumes M: "m.mono_body F"
    assumes W: "wf V"
    assumes I: "P x"
    assumes S: "\<And>D x. \<lbrakk>P x; \<And>x'. \<lbrakk>P x'; (x', x)\<in>V\<rbrakk> \<Longrightarrow> D x' \<le> mspec (Q x')\<rbrakk> \<Longrightarrow> F D x \<le> mspec (Q x)"
    shows "m.rec F x \<le> mspec (Q x)"
  proof -
    have "P x \<longrightarrow> m.rec F x \<le> mspec (Q x)"  
      apply (rule m.basic_rec_wf_rule[OF M W])
      using S by blast
    with I show ?thesis by blast
  qed
  
  lemma spec_m_let:
    assumes "f a \<le> mspec P"
    shows "let x=a in f x \<le> mspec P"
    using assms by auto
  
  lemma spec_m_if:
    assumes "b \<Longrightarrow> m\<^sub>1 \<le> mspec P"
    assumes "\<not>b \<Longrightarrow> m\<^sub>2 \<le> mspec P"
    shows "(if b then m\<^sub>1 else m\<^sub>2) \<le> mspec P"  
    using assms by auto
    
  (* TODO: Automation for datatypes! *)  
  lemma spec_m_prod: 
    assumes "\<And>a b. p=(a,b) \<Longrightarrow> f a b \<le> mspec P"
    shows "(case p of (a,b) \<Rightarrow> f a b) \<le> mspec P" 
    using assms by (cases p) auto
  
  lemma mpw_leq_M_iff[mpw_init]: "mi \<le> m \<longleftrightarrow> (\<forall>s. run mi s \<le> run m s)" 
    unfolding less_eq_M_def by (simp add: le_fun_def)
  
  lemma mpw_eq_M_iff[mpw_init]: "m=m' \<longleftrightarrow> (\<forall>s. run m s = run m' s)"
    by (cases m; cases m'; auto) 
    
  lemma M_htriple_iff: "m \<le> spec P Q \<longleftrightarrow> (\<forall>s. P s \<longrightarrow> run m s \<le> mspec (\<lambda>(x,s). Q x s))"  
    by mpw
    
  lemma M_htriple'_iff: "m \<le> spec' P Q \<longleftrightarrow> (\<forall>xs s. P xs s \<longrightarrow> run m s \<le> mspec (\<lambda>(x,s). Q xs x s))"  
    by mpw
    
  lemma M_htripleI: 
    assumes "\<And>s. P s \<Longrightarrow> run m s \<le> mspec (\<lambda>(x,s). Q x s)"  
    shows "m \<le> spec P Q"
    using assms by (simp add: M_htriple_iff)

  lemma M_htripleI': 
    assumes "\<And>xs s. P xs s \<Longrightarrow> run m s \<le> mspec (\<lambda>(x,s). Q xs x s)"  
    shows "m \<le> spec' P Q"
    using assms by (simp add: M_htriple'_iff)
    
        
  lemma spec_return:
    assumes "P (x,s)"
    shows "run (return x) s \<le> mspec P"  
    using assms by mpw
  
  lemma spec_bind:
    assumes "run m s \<le> mspec (\<lambda>(x,s). run (f x) s \<le> mspec P)"
    shows "run (doN { x\<leftarrow>m; f x }) s \<le> mspec P"  
    using assms
    by mpw'

  lemma spec_get:
    assumes "P (s,s)"
    shows "run get s \<le> mspec P"  
    using assms by mpw
    
  lemma spec_put:
    assumes "P ((),s)"  
    shows "run (put s) s' \<le> mspec P"
    using assms by mpw
    
        
  lemma spec_rec:
    assumes M: "mono_body F"
    assumes W: "wf V"
    assumes I: "P arb t x s"
    assumes S: "\<And>D arb t x s. \<lbrakk> P arb t x s; \<And>t' x' s'. \<lbrakk> P arb t' x' s'; (t',t)\<in>V \<rbrakk> \<Longrightarrow> run (D x') s' \<le> mspec (Q arb x' s')\<rbrakk> \<Longrightarrow> run (F D x) s \<le> mspec (Q arb x s)"
    shows "run (rec F x) s \<le> mspec (Q arb x s)"
  proof -
    have "\<forall>x s arb. P arb t x s \<longrightarrow> run (rec F x) s \<le> mspec (Q arb x s)"  
      apply (rule basic_rec_wf_rule[OF M W])
      using S by blast
    with I show ?thesis by blast  
  qed
  
  
  
  lemma spec_let:
    assumes "run (f a) s \<le> mspec P"
    shows "run (let x=a in f x) s \<le> mspec P"
    using assms by auto
  
  lemma spec_if:
    assumes "b \<Longrightarrow> run m\<^sub>1 s \<le> mspec P"
    assumes "\<not>b \<Longrightarrow> run m\<^sub>2 s \<le> mspec P"
    shows "run (if b then m\<^sub>1 else m\<^sub>2) s \<le> mspec P"  
    using assms by auto
    
  (* TODO: Automation for datatypes! *)  
  lemma spec_prod_case: 
    assumes "\<And>a b. p=(a,b) \<Longrightarrow> run (f a b) s \<le> mspec P"
    shows "run (case p of (a,b) \<Rightarrow> f a b) s \<le> mspec P" 
    using assms by (cases p) auto
  
  thm prod.split[where P="\<lambda>m. run m _ \<le> mspec _"]
  
  term spec
*)  
  
subsection \<open>Derived Combinators\<close>  

  definition IfM ("(if\<^sub>m (_)/ then (_)/ else (_))" [0, 0, 10] 10)
    where [rf2_mkpw]: "IfM P x y \<equiv> doN { b\<leftarrow>P; if b then x else y }"

    
  term PF_mono  
  lemma IfM_mono[partial_function_mono]: "monotone ord leq P \<Longrightarrow> monotone ord leq x \<Longrightarrow> monotone ord leq y \<Longrightarrow> monotone ord leq (\<lambda>D. IfM (P D) (x D) (y D))"
    unfolding IfM_def
    by pf_mono_prover  
    
    
  lemma gwp_IfM: "(gwp b s F (\<lambda>r s. if r then gwp t s F Q else gwp e s F Q)) \<Longrightarrow> gwp (if\<^sub>m b then t else e) s F Q"  
    by (mpws metis) 

  text \<open>Only holds for \<open>F=top\<close>\<close>
  lemma cgwp_IfM: "(cgwp b s top (\<lambda>r s. if r then cgwp t s top Q else cgwp e s top Q)) \<Longrightarrow> cgwp (if\<^sub>m b then t else e) s top Q"  
    by (mpws metis) 

  (*  
  lemma refine_IfM:
    assumes "refinec bi b si s (\<lambda>bi b si s. refinec (if bi then ti else ei) (if b then t else e) si s Q)"
    shows "refinec (if\<^sub>m bi then ti else ei) (if\<^sub>m b then t else e) si s Q"  
    unfolding IfM_def
    apply (rule refine_bind)
    by fact
  *)  
    
    
  definition "assert P \<equiv> if\<^sub>m get\<bind>return o P then return () else fail"

  lemma mpw_assert[mpw_simp]:
    "ismfail (run (assert P) s) \<longleftrightarrow> \<not>P s"
    "ismres (run (assert P) s) r i \<longleftrightarrow> (P s \<longrightarrow> r=((),s) \<and> i=0)"
    unfolding assert_def
    apply (force simp: mpw_simp)
    apply (auto simp: mpw_simp)
    done
  
  
  lemma gwp_assert: "gwp (assert P) s F Q \<longleftrightarrow> (F \<longrightarrow> P s \<longrightarrow> Q () s) \<and> (\<not>F \<longrightarrow> P s \<and> Q () s)"
    by mpw
  
  lemma cgwp_assert: "cgwp (assert P) s F Q \<longleftrightarrow> (F \<longrightarrow> P s \<longrightarrow> Q () s) \<and> (\<not>F \<longrightarrow> P s \<and> Q () s)"
    by mpw
  
  (*
  text \<open>Symmetric assertion refinement rule.\<close>  
  lemma refine_assert:
    assumes "P s \<Longrightarrow> Pi si"
    assumes "\<lbrakk> P s; Pi si \<rbrakk> \<Longrightarrow> Q () () si s"
    shows "refinec (assert Pi) (assert P) si s Q"
    using assms by m pw
  *)  
    
  definition "WhileM b c \<equiv> rec (\<lambda>while \<sigma>. if\<^sub>m b \<sigma> then doN {\<sigma>\<leftarrow>c \<sigma>; while \<sigma>} else return \<sigma>)"

  lemma WhileM_mono[partial_function_mono]: "\<lbrakk> \<And>\<sigma>. monotone ord leq (\<lambda>D. b \<sigma> D); \<And>\<sigma>. monotone ord leq (\<lambda>D. c \<sigma> D) \<rbrakk> \<Longrightarrow> monotone ord leq (\<lambda>D. WhileM (\<lambda>\<sigma>. b \<sigma> D) (\<lambda>\<sigma>. c \<sigma> D) \<sigma>)"
    unfolding WhileM_def
    by pf_mono_prover  
  
  lemma WhileM_unfold: "WhileM b c \<sigma> = (if\<^sub>m b \<sigma> then doN {\<sigma>\<leftarrow>c \<sigma>; WhileM b c \<sigma>} else return \<sigma>)"
    unfolding WhileM_def
    apply (subst rec_unfold)  
    apply pf_mono_prover
    ..
      
  lemma gwp_WhileM:
    fixes V :: "'t rel" and t :: 't
    assumes WF: "wf V"
    assumes I0: "I t \<sigma> s"
    assumes S: "\<And>t \<sigma> s. \<lbrakk>I t \<sigma> s\<rbrakk> \<Longrightarrow> gwp (b \<sigma>) s F (\<lambda>r s. 
        (r \<longrightarrow> gwp (c \<sigma>) s F (\<lambda>x' s'. \<exists>t'. (t',t)\<in>V \<and> I t' x' s' ))
      \<and> (\<not>r \<longrightarrow> Q \<sigma> s)  
      )"
    shows "gwp (WhileM b c \<sigma>) s F Q"  
    unfolding WhileM_def
    apply (rule gwp_rec[OF _ WF, where P="\<lambda>_::unit. I"])
    apply pf_mono_prover
    apply fact
    apply (rule gwp_IfM)
    apply (erule (1) gwp_conseq[OF S])
    apply mpw
    done
    

  lemma cgwp_WhileM:
    fixes V :: "'t rel" and t :: 't
    assumes WF: "wf V"
    assumes I0: "I t \<sigma> s"
    assumes S: "\<And>t \<sigma> s. \<lbrakk>I t \<sigma> s\<rbrakk> \<Longrightarrow> cgwp (b \<sigma>) s top (\<lambda>r s. 
        (r \<longrightarrow> cgwp (c \<sigma>) s F (\<lambda>x' s'. \<exists>t'. (t',t)\<in>V \<and> I t' x' s' ))
      \<and> (\<not>r \<longrightarrow> Q \<sigma> s)  
      )"
    shows "cgwp (WhileM b c \<sigma>) s top Q"  
    unfolding WhileM_def
    apply (rule cgwp_rec[OF _ WF, where P="\<lambda>_::unit. I"])
    apply pf_mono_prover
    apply fact
    apply (rule cgwp_IfM)
    apply (erule (1) cgwp_conseq[OF S])
    apply (mpws metis) 
    done
    
      
  (* TODO: Move *)
  notation rel_prod (infixr "\<times>\<^sub>r" 60)  


  xxx, ctd here: how to encode parallelism? 
    do we need an extension of monad
  
  
      
subsection \<open>Instantiation to Separation Logic\<close>

definition assn 
  :: "('ai \<Rightarrow>'a \<Rightarrow> 'asi::stronger_sep_algebra \<times> 'as::stronger_sep_algebra \<Rightarrow> bool) \<Rightarrow> 'ai \<Rightarrow> 'a \<Rightarrow> 'asi \<times> 'as \<Rightarrow> bool" ("\<upharpoonleft>_" [1000] 1000) 
  where "assn A xi x \<equiv> A xi x"

definition "pure A ai a \<equiv> \<up>(A ai a)"  


context
  fixes \<alpha>i :: "'asi::stronger_sep_algebra \<Rightarrow> 'si \<Rightarrow> bool"
    and \<alpha> :: "'as::stronger_sep_algebra \<Rightarrow> 's \<Rightarrow> bool"
begin

  definition "RSTATE P si s \<equiv> \<exists>asi as. \<alpha>i asi si \<and> \<alpha> as s \<and> P (asi, as)"
  
  definition [mpw_init]: "srefine P Q mi m \<equiv> \<forall>si s F. 
    RSTATE (P**F) si s 
    \<longrightarrow> refinec mi m si s (\<lambda>ri r si s. RSTATE (Q ri r ** F) si s)"

  lemma RSTATE_conseq: "RSTATE P si s \<Longrightarrow> P \<turnstile> P' \<Longrightarrow> RSTATE P' si s"
    unfolding RSTATE_def by auto  
    
  lemma RSTATE_FRAME_INFER: "RSTATE P si s \<Longrightarrow> PRECOND (FRAME_INFER P P' F) \<Longrightarrow> RSTATE (P'**F) si s"
    by (meson FRAME_INFER_def RSTATE_conseq PRECOND_def)
    
  lemma srefine_srefinecI:
    assumes "srefine P Q mi m"
    assumes "RSTATE (P**F) si s"
    assumes "\<And>ri si r s. RSTATE (Q ri r ** F) si s \<Longrightarrow> C ri r si s"
    shows "refinec mi m si s C"
    using assms
    by mpw
    
  lemma srefinec_srefineI:
    assumes "\<And>si s F. RSTATE (P**F) si s \<Longrightarrow> refinec mi m si s (\<lambda>ri r si s. RSTATE (Q ri r ** F) si s)"
    shows "srefine P Q mi m"  
    using assms unfolding srefine_def
    by (simp add: )
    
  lemma srefinec_srefineI':
    assumes "\<And>si s F. RSTATE (P**EXACT F) si s \<Longrightarrow> refinec mi m si s (\<lambda>ri r si s. RSTATE (Q ri r ** EXACT F) si s)"
    shows "srefine P Q mi m"  
    using assms unfolding srefine_def
    apply mpwt
    apply (auto simp: RSTATE_def sep_conj_def sep_algebra_simps intro: sep_disj_commuteI)
    apply (blast intro: sep_disj_commuteI)
    by (smt (verit, ccfv_SIG) sep_disj_commuteI)
    
  (* Now assuming that operands are variables, being in relation *)  
    
  lemma srefine_return: "srefine (\<upharpoonleft>R xi x) (\<lambda>ri r. \<upharpoonleft>R ri r) (return xi) (return x)"  
    by mpw
   
  lemma srefine_rec:
    assumes P: "RSTATE (\<upharpoonleft>R xi x ** \<Gamma>) si s"
    assumes M: "mono_body Fi"
    assumes S: "\<And>xi si x s Di D. \<lbrakk> RSTATE (\<upharpoonleft>R xi x ** \<Gamma>) si s; \<And>xi si x s. RSTATE (\<upharpoonleft>R xi x ** \<Gamma>) si s \<Longrightarrow> refinec (Di xi) (D x) si s C \<rbrakk> \<Longrightarrow> refinec (Fi Di xi) (F D x) si s C"
    shows "refinec (rec Fi xi) (rec F x) si s C"
    using S
    apply -
    apply (rule refine_rec[OF M, where Ra="\<lambda>xi si x s. RSTATE (\<upharpoonleft>R xi x ** \<Gamma>) si s"])
    apply (rule P)
    by blast

  lemma srefine_if:
    assumes "VC (bi=b)"
    assumes "\<lbrakk> bi;b \<rbrakk> \<Longrightarrow> refinec ti t si s Q"
    assumes "\<lbrakk> \<not>bi;\<not>b \<rbrakk> \<Longrightarrow> refinec ei e si s Q"
    shows "refinec (if bi then ti else ei) (if b then t else e) si s Q"  
    using assms by auto
    
            
end

context
  fixes \<alpha> :: "'as::stronger_sep_algebra \<Rightarrow> 's \<Rightarrow> bool"
begin
  definition PURE :: "bool \<Rightarrow> bool" where [simp]: "PURE P = P"

  definition "STATE P s \<equiv> \<exists>as. \<alpha> as s \<and> P as"
    
  definition "POSTCOND P s \<equiv> STATE P s"
  
  
  definition [mpw_init]: "gen_htriple' P m X Q 
    \<equiv> \<forall>l s F. STATE (P l**F) s \<longrightarrow> gwp m s (X l s) (\<lambda>r s. STATE (Q l r ** F) s)"
  
  abbreviation "gen_htriple P m X Q \<equiv> gen_htriple' (\<lambda>_::unit. P) m X (\<lambda>_. Q)"
  
  lemma htripleD: 
    assumes "gen_htriple' P m X Q"
    assumes "STATE (P l**F) s"
    shows "gwp m s (X l s) (\<lambda>r s. POSTCOND (Q l r ** F) s)"
    using assms unfolding POSTCOND_def by mpw
  
  lemma STATE_conseq: "STATE P s \<Longrightarrow> P \<turnstile> Q \<Longrightarrow> STATE Q s"
    using STATE_def by auto

  lemma htripleD': 
    assumes "gen_htriple' P m X Q"
    assumes "STATE PP s"
    assumes "PP \<turnstile> P l ** F"
    assumes "X l s \<Longrightarrow> X'"
    assumes "\<And>r s'. STATE (Q l r ** F) s' \<Longrightarrow> QQ r s'"
    shows "gwp m s X' QQ"
    using assms STATE_conseq 
    by (smt (z3) POSTCOND_def gwp_conseq htripleD)
        
  lemma STATE_FRAME_INFER: "STATE P s \<Longrightarrow> PRECOND (FRAME_INFER P P' F) \<Longrightarrow> STATE (P'**F) s"
    by (meson FRAME_INFER_def STATE_conseq PRECOND_def)
    
  lemma gen_htripleI:  
    assumes "\<And>l s F. STATE (P l**F) s \<Longrightarrow> gwp m s (X l s) (\<lambda>r. POSTCOND (Q l r ** F))"
    shows "gen_htriple' P m X Q"
    using assms unfolding POSTCOND_def
    by mpw
    
  lemma gen_htripleI':  
    assumes "\<And>l s F. STATE (P l ** EXACT F) s \<Longrightarrow> gwp m s (X l s) (\<lambda>r. POSTCOND (Q l r ** EXACT F))"
    shows "gen_htriple' P m X Q"
    using assms unfolding POSTCOND_def STATE_def
    apply mpwt
    apply (auto simp: STATE_def[abs_def] sep_algebra_simps sep_conj_def)
    using sep_disj_commuteI apply blast
    by (metis (no_types, hide_lams) sep_disj_commuteI)
  
  lemma htriple_sspecI: 
    assumes "gen_htriple' P m X Q"
    assumes "STATE (P l**F) s"
    assumes "\<And>r s. STATE (Q l r ** F) s \<Longrightarrow> Q' r s"
    assumes "X l s \<Longrightarrow> X'"
    shows "gwp m s X' (\<lambda>r s. Q' r s)"
    using assms by mpw' 

  (* return-pass version *)      
  lemma   
    shows "gen_htriple (A x) (return x) bot (\<lambda>r. A r)"
    by mpw
    
  lemma sspec_return: "gen_htriple P (return x) bot (\<lambda>r. \<up>(r=x) ** P)" 
    supply [simp] = sep_algebra_simps
    by mpw'
        
  lemma gen_htriple_extract_pure: "gen_htriple' (\<lambda>l. \<up>\<phi> ** P l) c X Q \<longleftrightarrow> (\<phi> \<longrightarrow> gen_htriple' P c X Q)"
    apply (cases \<phi>)
    apply (auto simp: sep_algebra_simps) []
    unfolding gen_htriple'_def STATE_def
    apply (auto)
    done
    
  lemma sspec_rec:
    assumes M: "mono_body F"
    assumes W: "wf V"
    assumes S: "\<And>D t x. \<lbrakk>\<And>t' x. gen_htriple' (\<lambda>l. P t' x l ** \<up>((t',t)\<in>V)) (D x) (X x) (\<lambda>l. Q l x)\<rbrakk> \<Longrightarrow> gen_htriple' (P t x) (F D x) (X x) (\<lambda>l. Q l x)"
    shows "gen_htriple' (P t x) (rec F x) (X x) (\<lambda>l. Q l x)"  
  proof -
    have "\<forall>x. gen_htriple' (P t x) (rec F x) (X x) (\<lambda>l. Q l x)"  
      apply (rule basic_rec_wf_rule[OF M W])
      apply (intro allI)
      apply (rule S)
      apply (simp add: sep_algebra_simps gen_htriple_extract_pure) 
      done
    then show ?thesis by blast  
  qed
    
  lemma sspec_rec':
    assumes "STATE (P arb t x) s"
    assumes M: "mono_body F"
    assumes W: "wf V"
    assumes "\<And>arb s t x D FF. \<lbrakk> STATE (P arb t x ** FF) s; \<And>s' t' x'. \<lbrakk> STATE (P arb t' x' ** \<up>((t',t)\<in>V) ** FF) s' \<rbrakk> \<Longrightarrow> gwp (D x') s' X C \<rbrakk> \<Longrightarrow> gwp (F D x) s X C"
    shows "gwp (rec F x) s X C"
    using assms
    apply -
    supply R = gwp_rec[OF M W, where P="\<lambda>(arb,FF) t x s. STATE (P arb t x ** FF) s" and arb="(arb,\<box>)"]
    apply (rule R)
    apply fastforce
    
    supply [simp] = sep_algebra_simps STATE_def
    apply clarsimp
    apply (rprems)
    apply auto
    done
    
  lemma sspec_let:
    assumes "\<And>x. PURE (x=a) \<Longrightarrow> gwp (f x) s X C"
    shows "gwp (let x=a in f x) s X C"
    using assms by auto
  
  lemma sspec_if:
    assumes "PURE b \<Longrightarrow> gwp m\<^sub>1 s X C"
    assumes "PURE (\<not>b) \<Longrightarrow> gwp m\<^sub>2 s X C"
    shows "gwp (if b then m\<^sub>1 else m\<^sub>2) s X C"  
    using assms by auto
    
  (* TODO: Automation for datatypes! *)  
  lemma sspec_prod_case: 
    assumes "\<And>a b. p=(a,b) \<Longrightarrow> gwp (f a b) s X C"
    shows "gwp (case p of (a,b) \<Rightarrow> f a b) s X C" 
    using assms by (cases p) auto
    
end


subsubsection \<open>Refinement Assertions\<close>

definition CONC :: "('a \<Rightarrow> bool) \<Rightarrow> 'a::sep_algebra \<times> 'b::sep_algebra \<Rightarrow> bool" 
  where "CONC P \<equiv> \<lambda>(c,a). P c \<and> a=0"
definition ABS :: "('b \<Rightarrow> bool) \<Rightarrow> 'a::sep_algebra \<times> 'b::sep_algebra \<Rightarrow> bool" 
  where "ABS P \<equiv> \<lambda>(c,a). c=0 \<and> P a"

definition PROJ_ABS :: "('a::sep_algebra \<times> 'b::sep_algebra \<Rightarrow> bool) \<Rightarrow> 'b \<Rightarrow> bool" 
  where "PROJ_ABS P s \<equiv> \<exists>si. P (si,s)"

definition PROJ_CONC :: "('a::sep_algebra \<times> 'b::sep_algebra \<Rightarrow> bool) \<Rightarrow> 'a \<Rightarrow> bool" 
  where "PROJ_CONC P si \<equiv> \<exists>s. P (si,s)"
  
lemma proj_abs_inv[sep_algebra_simps]: "PROJ_ABS (ABS P) = P"
  unfolding PROJ_ABS_def ABS_def by auto
  
lemma proj_conc_inv[sep_algebra_simps]: "PROJ_CONC (CONC P) = P"
  unfolding PROJ_CONC_def CONC_def by auto
      
lemma proj_abs_conc[sep_algebra_simps]: "PROJ_ABS (CONC P) = (\<up>(P\<noteq>sep_false))"  
  unfolding PROJ_CONC_def CONC_def PROJ_ABS_def
  by (auto simp: sep_algebra_simps fun_eq_iff)
  
lemma proj_conc_abs[sep_algebra_simps]: "PROJ_CONC (ABS P) = (\<up>(P\<noteq>sep_false))"  
  unfolding PROJ_CONC_def CONC_def PROJ_ABS_def ABS_def
  by (auto simp: sep_algebra_simps fun_eq_iff)
  
lemma abs_star[sep_algebra_simps]: "ABS (A**B) = (ABS A ** ABS B)"
  by (auto simp: ABS_def sep_conj_def fun_eq_iff sep_algebra_simps)
  
lemma conc_star[sep_algebra_simps]: "CONC (A**B) = (CONC A ** CONC B)"
  by (auto simp: CONC_def sep_conj_def fun_eq_iff sep_algebra_simps)
  
lemma PROJ_ABS_star: "PROJ_ABS (A**B) \<turnstile> (PROJ_ABS A ** PROJ_ABS B)"
  by (auto 0 4 simp: PROJ_ABS_def sep_conj_def fun_eq_iff sep_algebra_simps)
  
lemma PROJ_CONC_star: "PROJ_CONC (A**B) \<turnstile> (PROJ_CONC A ** PROJ_CONC B)"
  by (auto 0 4 simp: PROJ_CONC_def sep_conj_def fun_eq_iff sep_algebra_simps)
      
subsubsection \<open>Proving refinement via Hoare-Style rules\<close>  
(*
  xxx, ctd here:
    idea: let both programs run, and show that results must be in relation?
    does this work with non-determisnism?
    
  


  definition CONC :: "('a \<Rightarrow> bool) \<Rightarrow> 'a::sep_algebra \<times> 'b::sep_algebra \<Rightarrow> bool" 
    where "CONC P \<equiv> \<lambda>(c,a). P c \<and> a=0"
  definition ABS :: "('b \<Rightarrow> bool) \<Rightarrow> 'a::sep_algebra \<times> 'b::sep_algebra \<Rightarrow> bool" 
    where "ABS P \<equiv> \<lambda>(c,a). c=0 \<and> P a"
  
  definition PROJ_ABS :: "('a::sep_algebra \<times> 'b::sep_algebra \<Rightarrow> bool) \<Rightarrow> 'b \<Rightarrow> bool" 
    where "PROJ_ABS P s \<equiv> \<exists>si. P (si,s)"

  definition PROJ_CONC :: "('a::sep_algebra \<times> 'b::sep_algebra \<Rightarrow> bool) \<Rightarrow> 'a \<Rightarrow> bool" 
    where "PROJ_CONC P si \<equiv> \<exists>s. P (si,s)"
    
  lemma "PROJ_ABS (ABS P) = P"
    unfolding PROJ_ABS_def ABS_def by auto
    
  lemma "PROJ_CONC (CONC P) = P"
    unfolding PROJ_CONC_def CONC_def by auto
        
  lemma "PROJ_ABS (CONC P) = (\<up>(P\<noteq>sep_false))"  
    unfolding PROJ_CONC_def CONC_def PROJ_ABS_def
    by (auto simp: sep_algebra_simps fun_eq_iff)
    
  lemma "PROJ_CONC (ABS P) = (\<up>(P\<noteq>sep_false))"  
    unfolding PROJ_CONC_def CONC_def PROJ_ABS_def ABS_def
    by (auto simp: sep_algebra_simps fun_eq_iff)
    
  lemma "ABS (A**B) = (ABS A ** ABS B)"
    by (auto simp: ABS_def sep_conj_def fun_eq_iff sep_algebra_simps)
    
  lemma "CONC (A**B) = (CONC A ** CONC B)"
    by (auto simp: CONC_def sep_conj_def fun_eq_iff sep_algebra_simps)
    
  lemma PROJ_ABS_star: "PROJ_ABS (A**B) \<turnstile> (PROJ_ABS A ** PROJ_ABS B)"
    by (auto 0 4 simp: PROJ_ABS_def sep_conj_def fun_eq_iff sep_algebra_simps)
    
  lemma PROJ_CONC_star: "PROJ_CONC (A**B) \<turnstile> (PROJ_CONC A ** PROJ_CONC B)"
    by (auto 0 4 simp: PROJ_CONC_def sep_conj_def fun_eq_iff sep_algebra_simps)
        
  lemma RSTATE_absD: "RSTATE \<alpha>i \<alpha> P si s \<Longrightarrow> STATE (PROJ_ABS P) s"  
    unfolding RSTATE_def STATE_def PROJ_ABS_def
    by auto

end

lemma RSTATE_concD: "RSTATE \<alpha>i \<alpha> P si s \<Longrightarrow> STATE \<alpha>i (PROJ_CONC P) si"  
  unfolding RSTATE_def STATE_def PROJ_CONC_def
  by auto
    
        
  find_theorems "rel_prod"
  
  lemma "a\<le>b \<Longrightarrow> (a**c) \<le> (b**c)" for a b c :: "_::sep_algebra \<Rightarrow> _"
    by (simp add: sep_rule(3))
    
  
  (* TODO: Can be done for any lens *)  
  definition "zoom_conc m \<equiv> M (\<lambda>(s\<^sub>1,s\<^sub>2). case run m s\<^sub>1 of mfail \<Rightarrow> mfail | mspec Q \<Rightarrow> mspec (\<lambda>(r,(s\<^sub>1',s\<^sub>2')). Q (r,s\<^sub>1') \<and> s\<^sub>2'=s\<^sub>2))"
  definition "zoom_abs m \<equiv> M (\<lambda>(s\<^sub>1,s\<^sub>2). case run m s\<^sub>2 of mfail \<Rightarrow> mfail | mspec Q \<Rightarrow> mspec (\<lambda>(r,(s\<^sub>1',s\<^sub>2')). Q (r,s\<^sub>2') \<and> s\<^sub>1'=s\<^sub>1))"
    
  lemma zoom_conc_mpw[mpw_simp]:
    "ismfail (run (zoom_conc m) s) \<longleftrightarrow> (case s of (s\<^sub>1,s\<^sub>2) \<Rightarrow> ismfail (run m s\<^sub>1))"
    "ismres (run (zoom_conc m) s) xs \<longleftrightarrow> (case s of (s\<^sub>1,s\<^sub>2) \<Rightarrow> case xs of (r,(s\<^sub>1',s\<^sub>2')) \<Rightarrow> ismfail (run m s\<^sub>1) \<or> ismres (run m s\<^sub>1) (r,s\<^sub>1') \<and> s\<^sub>2'=s\<^sub>2 )"
    unfolding zoom_conc_def 
    supply [split] = M.splits mres.splits
    by mpw'+
    
  lemma zoom_abs_mpw[mpw_simp]:
    "ismfail (run (zoom_abs m) s) \<longleftrightarrow> (case s of (s\<^sub>1,s\<^sub>2) \<Rightarrow> ismfail (run m s\<^sub>2))"
    "ismres (run (zoom_abs m) s) xs \<longleftrightarrow> (case s of (s\<^sub>1,s\<^sub>2) \<Rightarrow> case xs of (r,(s\<^sub>1',s\<^sub>2')) \<Rightarrow> ismfail (run m s\<^sub>2) \<or> ismres (run m s\<^sub>2) (r,s\<^sub>2') \<and> s\<^sub>1'=s\<^sub>1 )"
    unfolding zoom_abs_def 
    supply [split] = M.splits mres.splits
    by mpw'+
  
  
  thm mpw_simp
  
  lemma 
    assumes "gen_htriple \<alpha>i P m Q"  
    shows "gen_htriple (rel_prod \<alpha>i \<alpha>) (CONC P) (zoom_conc m) (\<lambda>r. CONC (Q r))"
    using assms
    unfolding gen_htriple'_def STATE_def CONC_def
    apply (auto)
    apply mpw't
    
    
    oops
    xxx, ctd here:
      this might be the right way to execute the concrete and abstract program one after the other,
      using their specifications to prove the refinement.
    
    

  lemma     
    fixes m :: "('a,'s) M" and  mi :: "('ai,'si) M"
    assumes NF: "\<And>s F. \<lbrakk> \<not>ismfail (run m s); STATE \<alpha> (PROJ_ABS (P**F)) s \<rbrakk> \<Longrightarrow> run m s \<noteq> mspec bot"
    assumes ABS_HT: "gen_htriple \<alpha> PP m QQ" 
    assumes CONC_HT: "gen_htriple \<alpha>i PPi mi QQi" 
    assumes ABS_ENT: "PROJ_ABS P \<turnstile> PP"
    assumes CONC_ENT: "PROJ_CONC P \<turnstile> PPi"
    assumes R_END: "\<And>ri r. CONC (QQi ri) ** ABS (QQ r) \<turnstile> Q ri r"
    shows "srefine \<alpha>i \<alpha> P Q mi m"
    unfolding srefine_def
    apply clarsimp
    subgoal for si s F
      apply (rule refine_by_specI[where Q="\<lambda>(r,s'). STATE \<alpha> (QQ r ** PROJ_ABS F) s'"]) 
      subgoal
        apply (rule NF, assumption)
        apply (simp add: RSTATE_def STATE_def PROJ_ABS_def; blast) 
        done
      subgoal
        apply (rule htripleD[OF ABS_HT])
        apply (drule RSTATE_absD)
        apply (drule STATE_conseq[OF _])
        apply (sep_drule PROJ_ABS_star)
        apply (sep_drule ABS_ENT)
        apply (rule order_refl)
        .
      subgoal for r
        apply (rule htripleD'[OF CONC_HT])
        apply (drule RSTATE_concD, assumption)
        apply (sep_drule PROJ_CONC_star)
        apply (sep_drule CONC_ENT)
        apply (rule order_refl)
        apply (cases r; simp)
        xxx, this lifting does not work, as coherence between abs and conc F is lost!
        
        
        
        using ABS_HT ABS_ENT 
      
    try to lift the Hoare-triples, instead of projecting the state!
        
        
    
    using assms
    unfolding STATE_def RSTATE_def PROJ_CONC_def PROJ_ABS_def POSTCOND_def CONC_def ABS_def
      gen_htriple'_def
    apply mpwt
    unfolding STATE_def RSTATE_def PROJ_CONC_def PROJ_ABS_def POSTCOND_def CONC_def ABS_def
    apply auto
    
    
    oops
    apply (rule refine_by_specI) 
    apply (rule NF, assumption, blast dest: RSTATE_absD[of \<alpha>]) 
    
  
*)    

subsection \<open>VCG\<close>
subsubsection \<open>Aux\<close>

text \<open>Solve during frame inference by reflexivity. 
  Used for, e.g., measure variables in invariants
\<close>
definition set_mvar (infix \<open>:=\<^sub>\<up>\<close> 50) where "set_mvar t x \<equiv> \<up>(t=x)"
lemma set_mvar_refl[fri_rules]: "\<box> \<turnstile> x :=\<^sub>\<up> x"
  unfolding set_mvar_def by (simp add: sep_algebra_simps)

  

subsubsection \<open>Normalizer\<close>

  lemma imp_eq_meta_imp: "Trueprop (a \<longrightarrow> b) \<equiv> (a \<Longrightarrow> b)" by presburger

  lemmas norm_conj_imp_simps = conj_imp_eq_imp_imp imp_eq_meta_imp
  
  lemmas norm_metaE = exE conjE impE
  lemmas norm_metaI = impI allI
  
  
  text \<open>Splitting Products\<close>
  named_simpset vcg_split_prod = HOL_basic_ss_nmp
  
  lemmas [named_ss vcg_split_prod] = split norm_conj_imp_simps case_unit_Unity
  lemma case_prod_merge[named_ss vcg_split_prod]:
    "((case p of (a,b) \<Rightarrow> P a b) ** Q) = (case p of (a,b) \<Rightarrow> P a b ** Q)"
    "(Q ** (case p of (a,b) \<Rightarrow> P a b)) = (Q ** (case p of (a,b) \<Rightarrow> P a b))"
    by auto
  
  lemma STATE_case_prod_extract[named_ss vcg_split_prod]: 
    "STATE \<alpha> (case p of (a,b) \<Rightarrow> P a b) s = (\<exists>a b. p=(a,b) \<and> STATE \<alpha> (P a b) s)"
    by (cases p) auto
  
  definition [simp]: "SPLIT p Q \<equiv> \<forall>a b. p=(a,b) \<longrightarrow> Q a b"  
    
  lemma SPLIT_simps[named_ss vcg_split_prod]:
    "\<And>Q. (\<exists>x. SPLIT p (\<lambda>a b. Q x a b)) \<longleftrightarrow> SPLIT p (\<lambda>a b. \<exists>x. Q x a b)"
    "\<And>P Q. (P \<longrightarrow> SPLIT p (\<lambda>a b. Q a b)) \<longleftrightarrow> SPLIT p (\<lambda>a b. P \<longrightarrow> Q a b)"
    by auto
  
  lemma SPLITI:  
    "\<lbrakk>\<And>a b. p=(a,b) \<Longrightarrow> Q a b\<rbrakk> \<Longrightarrow> SPLIT p (\<lambda>a b. Q a b)"
    by auto
  
  lemma POSTCOND_case_prod_extract[named_ss vcg_split_prod]:
    "POSTCOND \<alpha> (case p of (a,b) \<Rightarrow> P a b) s = (SPLIT p (\<lambda>a b. POSTCOND \<alpha> (P a b) s))"
    by (cases p) auto
  
  method vcg_split_prod = (changed \<open>
    (simp named_ss vcg_split_prod:)?,
    (elim norm_metaE)?; (intro norm_metaI SPLITI)?;
    hypsubst?
    \<close>)+
    
  

  named_simpset vcg_extract_pure = HOL_basic_ss_nmp

  definition "PURE_extr P \<equiv> \<up>P"

  lemmas [named_ss vcg_extract_pure] = sep_conj_assoc sep_conj_i
  lemmas [named_ss vcg_extract_pure] = set_mvar_def
  
  lemma pred_lift_PURE_simps[named_ss vcg_extract_pure]:
    "\<up>P = PURE_extr P"
    "NO_MATCH (PURE_extr X) A \<Longrightarrow> (A ** PURE_extr \<Phi> ** B) = (PURE_extr \<Phi> ** A ** B)"
    "NO_MATCH (PURE_extr X) A \<Longrightarrow> (A ** PURE_extr \<Phi>) = (PURE_extr \<Phi> ** A)"
    by (auto simp: pred_lift_def sep_conj_def sep_algebra_simps PURE_extr_def)
    
  lemma PURE_extrE:
    "STATE \<alpha> (PURE_extr \<Phi> ** P) s \<Longrightarrow> \<lbrakk> PURE \<Phi> \<Longrightarrow> STATE \<alpha> P s \<Longrightarrow> thesis \<rbrakk> \<Longrightarrow> thesis"
    "STATE \<alpha> (PURE_extr \<Phi>) s \<Longrightarrow> \<lbrakk> PURE \<Phi> \<Longrightarrow> STATE \<alpha> \<box> s \<Longrightarrow> thesis \<rbrakk> \<Longrightarrow> thesis"
    unfolding PURE_extr_def STATE_def by (auto simp: sep_algebra_simps)
  
  lemma STATE_assn_cong: "P\<equiv>Q \<Longrightarrow> STATE \<alpha> P s \<equiv> STATE \<alpha> Q s" by simp
    
  
  ML \<open>
    structure VCG_Extract_Pure = struct
    
      fun STATE_conv ctxt = RF2_Util.cong_rls_conv 
        @{thms STATE_assn_cong} 
        (Named_Simpsets.rewrite @{named_simpset vcg_extract_pure} [] ctxt)
    
      fun tac ctxt = 
        CHANGED o CONVERSION (Conv.top_sweep_conv STATE_conv ctxt)
        THEN' TRY o elim_determ_tac ctxt @{thms PURE_extrE}
  
    
    end
  \<close>    
    
  method_setup vcg_extract_pure = \<open>Scan.succeed (SIMPLE_METHOD' o VCG_Extract_Pure.tac)\<close>
      

  (*        
  lemma POSTCOND_cong(*[named_ss vcg_norm cong]*): "POSTCOND \<alpha> P s = POSTCOND \<alpha> P s" by simp
  lemma FRAME_INFER_cong: "P=P' \<Longrightarrow> FRAME_INFER P Q F = FRAME_INFER P' Q F" by simp
  
  
  lemma STATE_extract_pure[named_ss vcg_extract_pure]:
    "STATE \<alpha> (\<up>\<phi>) s \<longleftrightarrow> PURE \<phi> \<and> STATE \<alpha> \<box> s"
    "STATE \<alpha> (\<up>\<phi> ** F) s \<longleftrightarrow> PURE \<phi> \<and> STATE \<alpha> F s"
    unfolding STATE_def by (auto simp: sep_algebra_simps)

  lemma FRAME_INFER_extract_pure[named_ss vcg_extract_pure]:
    "FRAME_INFER (\<up>\<phi>) Q F \<longleftrightarrow> (PURE \<phi> \<longrightarrow> FRAME_INFER \<box> Q F)"
    "FRAME_INFER (\<up>\<phi> ** P) Q F \<longleftrightarrow> (PURE \<phi> \<longrightarrow> FRAME_INFER P Q F)"
    unfolding FRAME_INFER_def 
    by (auto simp: sep_algebra_simps)
    
  lemmas [named_ss vcg_extract_pure] = pred_lift_move_front_simps
  lemmas [named_ss vcg_norm] = sep_conj_assoc sep_conj_i pure_true_conv
  
  lemmas [named_ss vcg_norm] = set_mvar_def
  *)
  
  method vcg_norm_goal = intro allI impI exI
  
  method vcg_norm = 
    (changed \<open>vcg_split_prod?, vcg_extract_pure?, vcg_norm_goal?\<close>)+



subsubsection \<open>Generator\<close>

definition [simp]: "RAW \<Phi> \<equiv> \<Phi>"

lemma RAWI: "\<Phi> \<Longrightarrow> RAW \<Phi>" by simp
lemma RAWD: "RAW \<Phi> \<Longrightarrow> \<Phi>" by simp

lemma FRAME_INFER_triv: "PRECOND (FRAME_INFER P P \<box>)"
  by (auto simp: PRECOND_def FRAME_INFER_def)

  
lemmas vcg_rule_simps = bot_apply bot_bool_def top_apply top_bool_def
  implies_True_equals False_implies_equals implies_False_swap True_implies_equals
  not_False_eq_True

named_theorems vcg_specialize  
    
lemma gwp_bot_specialize[vcg_specialize]: "gwp m x False Q \<Longrightarrow> gwp m x False Q" .
lemma cgwp_top_specialize[vcg_specialize]: "cgwp m x True Q \<Longrightarrow> cgwp m x True Q" .
  
  
ML \<open>
  structure VCG = struct
  
    fun try_first _ [] = NONE
      | try_first f (x::xs) = case try f x of SOME r => SOME r | NONE => try_first f xs
  
  
    val gen_thms = @{thms htriple_sspecI srefine_srefinecI}
    val frame_thms = @{thms STATE_FRAME_INFER RSTATE_FRAME_INFER}
  
    fun apply_thm thm fthm = thm OF [fthm]
    fun apply_thms fthms thm = the_default thm (try_first (apply_thm thm) fthms)
    fun apply_thms' fthms ctxt = singleton (Variable.trade (K (map (apply_thms fthms))) ctxt)
    
    fun apply_rev_thm thm fthm = fthm OF [thm]
    fun apply_rev_thms fthms thm = the_default thm (try_first (apply_rev_thm thm) fthms)
    fun apply_rev_thms' fthms ctxt = singleton (Variable.trade (K (map (apply_rev_thms fthms))) ctxt)

    fun vcg_thm_conv_iff thm = let
      fun aux thm = thm RS @{thm iffD2}
    
    in
      perhaps (try aux) thm
    end
    
    fun vcg_thm_conv1 ctxt thm =
      case try (fn thm => thm RS @{thm RAWD}) thm of
        SOME thm => thm
      | NONE => (apply_rev_thms' gen_thms ctxt #> apply_thms' frame_thms ctxt) thm  
    
    fun remove_triv_frame_infer ctxt thm = let
      
      val pvars = case Thm.prems_of thm of
        (t as @{mprop "STATE _ _ _"}) :: _ => Term.add_vars t []
      | _ => []  
      val vars = Term.add_vars (Thm.concl_of thm) pvars |> map Var
      
      fun ipr @{mprop \<open>PRECOND (FRAME_INFER _ ?Q ?F)\<close>} =
        if is_Var Q andalso is_Var F andalso not (member op= vars Q) then @{thm FRAME_INFER_triv} else asm_rl
      | ipr _ = asm_rl  
      
    
      val insts = map ipr (Thm.prems_of thm)
      
      val thm = (thm OF insts) |> RF2_Util.simplified @{thms sep_conj_i} ctxt
    in thm end  
      
    fun vcg_thm_conv2 ctxt = 
      vcg_thm_conv_iff 
      #> vcg_thm_conv1 ctxt 
      #> RF2_Util.simplified @{thms vcg_rule_simps} ctxt
      #> remove_triv_frame_infer ctxt
      #> Drule.zero_var_indexes

    fun vcg_thm_specialize1 ctxt thm sthm = case try (fn sthm => thm RS sthm) sthm of
      NONE => NONE
    | SOME thm' => 
        if Thm.eq_thm (thm', thm) then NONE
        else let
          val thm' = vcg_thm_conv2 ctxt thm'
        
        in SOME thm'
        end
  
  
    fun vcg_thm_specialize ctxt thm = 
      let
        val sthms = Named_Theorems.get ctxt @{named_theorems vcg_specialize}
      in 
        map_filter (vcg_thm_specialize1 ctxt thm) sthms @ [thm]
      end

      
    fun vcg_thm_conv ctxt = 
      vcg_thm_conv2 ctxt
      #> vcg_thm_specialize ctxt
            
      
    fun has_state thm = case Thm.prems_of thm of (@{mprop \<open>STATE _ _ _\<close>} :: _) => true | _ => false
    fun state_rl_tac ctxt thm = if has_state thm then eresolve_tac ctxt [thm] else resolve_tac ctxt [thm]
    fun state_rls_tac ctxt thms = APPEND_LIST' (map (state_rl_tac ctxt) thms)  
    
    (* TODO: Net versions of this tactic! *)
    
  end

\<close>


attribute_setup vcg_xformed = 
  \<open>Scan.succeed (Thm.rule_attribute [] (fn context => 
    VCG.vcg_thm_conv2 (Context.proof_of context)
  ))\<close>
  \<open>Apply static rule transformations\<close>

attribute_setup vcg_xformed_spec = 
  \<open>Scan.succeed (Thm.rule_attribute [] (fn context => 
    VCG.vcg_thm_conv (Context.proof_of context)
    #> Conjunction.intr_balanced
  ))\<close>
  \<open>Apply static rule transformations and specializations\<close>
  
  
named_theorems2 vcg_rules (reverse) morphisms \<open>VCG.vcg_thm_conv o Context.proof_of\<close> and \<open>K I\<close>
  
method_setup state_rule = \<open>Attrib.thms >> (fn thms => fn ctxt => SIMPLE_METHOD' (VCG.state_rls_tac ctxt thms))\<close>
method_setup vcg_rule' = \<open>Attrib.thms >> (fn thms => fn ctxt => let
  val thms = map (VCG.vcg_thm_conv ctxt) thms |> flat
in
  SIMPLE_METHOD' (VCG.state_rls_tac ctxt thms)
end  
  )\<close>
method_setup vcg_rule = \<open>Attrib.thms >> (fn thms => fn ctxt => let
  val thms = map (VCG.vcg_thm_conv ctxt) thms  |> flat
in
  SIMPLE_METHOD' (WITH_PRECOND (VCG.state_rls_tac ctxt thms) ctxt)
end  
  )\<close>

method vcg_step_nn = with_precond \<open>state_rule vcg_rules\<close>
method vcg_step = vcg_norm?; vcg_step_nn
method vcg' = repeat_all_new_fwd \<open>vcg_step\<close>
method vcg = prim_focus, vcg', prim_unfocus


subsection \<open>VCG Setup\<close>


lemmas [vcg_rules] = 
  srefinec_srefineI'[THEN RAWI]
  srefine_return refine_bind srefine_rec
  refine_let srefine_if refine_prod_case
  refine_IfM
  
  gen_htripleI'[THEN RAWI]
  sspec_return gwp_bind (*sspec_rec*)
  sspec_let sspec_if sspec_prod_case
  gwp_IfM

    
(*lemma sspec_IfM[vcg_rules]:
  assumes "gwp b s X (\<lambda>b s. gwp (if b then t else e) s X C)"
  shows "gwp (if\<^sub>m b then t else e) s X C"  
  using assms unfolding ss pecc_def by (ru le spec_IfM)
*)  

thm gwp_assert

lemma sspecc_assert[vcg_rules]:
  assumes "\<not>F \<Longrightarrow> VC (P s)"
  assumes "\<lbrakk>\<not>F; P s\<rbrakk> \<Longrightarrow> C () s"
  assumes "F \<Longrightarrow> P s \<Longrightarrow> C () s"
  shows "gwp (assert P) s F C"
  using assms unfolding gwp_assert
  by simp 

(* TODO: Assymetric bind-refine rules with sep-logic. for now: special case with assertions *)  
  
(*    
xxx: Two different modes for refinement proofs:

  automatic transfers: target is synthesized from source

  manual refinements proofs: both, target and source given
*)  
  
lemma srefine_WhileM:
  assumes A: "RSTATE \<alpha>i \<alpha> (\<upharpoonleft>Ra \<sigma>i \<sigma> ** \<Gamma>) si s"
  assumes B: "\<And>Di D \<sigma>i si \<sigma> s.
     \<lbrakk>\<And>\<sigma>i si \<sigma> s. RSTATE \<alpha>i \<alpha> (\<upharpoonleft>Ra \<sigma>i \<sigma> ** \<Gamma>) si s \<Longrightarrow> refinec (Di \<sigma>i) (D \<sigma>) si s Q; RSTATE \<alpha>i \<alpha> (\<upharpoonleft>Ra \<sigma>i \<sigma> ** \<Gamma>) si s\<rbrakk>
     \<Longrightarrow> refinec (if\<^sub>m bi \<sigma>i then ci \<sigma>i \<bind> Di else return \<sigma>i) (if\<^sub>m b \<sigma> then c \<sigma> \<bind> D else return \<sigma>) si s Q"
  shows "refinec (WhileM bi ci \<sigma>i) (WhileM b c \<sigma>) si s Q" 
  unfolding WhileM_def
  apply (rule srefine_rec)
  apply fact
  apply refine_mono
  apply (rule B)
  .

lemma sspec_WhileM:
  fixes V :: "'t rel" and t :: 't
  assumes I0: "STATE \<alpha> (I t \<sigma> ** \<Gamma>) s"
  assumes WF: "VC (wf V)"
  assumes S: "\<And>t \<sigma> s. \<lbrakk>STATE \<alpha> (I t \<sigma> ** \<Gamma>) s \<rbrakk> \<Longrightarrow> gwp (b \<sigma>) s X (\<lambda>r s. 
      (PURE r \<longrightarrow> gwp (c \<sigma>) s X (\<lambda>\<sigma>' s'. \<exists>t'. POSTCOND \<alpha> (I t' \<sigma>' ** \<up>((t',t)\<in>V) ** \<Gamma>) s'))
    \<and> (PURE (\<not>r) \<longrightarrow> C \<sigma> s)  
    )"
  (*assumes "\<And>t x s. \<lbrakk>I t x s\<rbrakk> \<Longrightarrow> run b s \<le> mspec (\<lambda>(r, s). run (if r then c x \<bind> D else return x) s \<le> mspec Q)"*)
  shows "gwp (WhileM b c \<sigma>) s X C"  
  apply (rule gwp_WhileM[where I="\<lambda>t \<sigma> s. STATE \<alpha> (I t \<sigma> ** \<Gamma>) s"])
  using WF apply simp
  using I0 apply blast
  apply (drule S)
  apply (simp add: sep_algebra_simps STATE_def POSTCOND_def) 
  apply mpw
  done
  
subsection \<open>Annotated Commands\<close>  

definition ASM :: "bool \<Rightarrow> bool" where [simp]: "ASM \<Phi> \<equiv> \<Phi>"
lemma ASMI: "\<Phi> \<Longrightarrow> ASM \<Phi>" by simp
declare_solver asm: ASMI = assumption
  

subsubsection \<open>Rec\<close>
  definition gen_recI :: "
      ('ca \<Rightarrow> 'c \<Rightarrow> bool) 
    \<Rightarrow> 't rel 
    \<Rightarrow> ('arb \<Rightarrow> 't \<Rightarrow> 'a \<Rightarrow> 'ca::stronger_sep_algebra \<Rightarrow> bool) 
    \<Rightarrow> ('arb \<Rightarrow> 'a \<Rightarrow> 'c \<Rightarrow> bool) 
    \<Rightarrow> ('arb \<Rightarrow> 'a \<Rightarrow> 'b \<Rightarrow> 'ca \<Rightarrow> bool) 
    \<Rightarrow> (('a \<Rightarrow> ('b, 'c) M) \<Rightarrow> 'a \<Rightarrow> ('b, 'c) M) 
    \<Rightarrow> 'a 
    \<Rightarrow> ('b, 'c) M" 
  where
    "gen_recI \<alpha> V P X Q \<equiv> rec"
  definition "RCALL f \<equiv> f"  

  lemma sspec_recI_aux:
    assumes M: "PRECOND (mono_body F)"
    assumes W: "VC (wf V)"
    assumes S: "\<And>D t x. \<lbrakk>\<And>t' x. gen_htriple' \<alpha> (\<lambda>l. P l t' x ** \<up>((t',t)\<in>V)) (D x) (\<lambda>l. X l x) (\<lambda>l. Q l x)\<rbrakk> \<Longrightarrow> gen_htriple' \<alpha> (\<lambda>l. P l t x) (F D x) (\<lambda>l. X l x) (\<lambda>l. Q l x)"
    shows "gen_htriple' \<alpha> (\<lambda>l. P l t x) (gen_recI \<alpha> V P X Q F x) (\<lambda>l. X l x) (\<lambda>l. Q l x)"  
    unfolding gen_recI_def
    apply (rule Monad.sspec_rec[where V=V])
    using assms
    unfolding PRECOND_def VC_pre_def
    apply auto 
    done
    
    
  lemma sspec_recI[THEN RAWI,vcg_rules]:
    assumes 
          "STATE \<alpha> PP s"
      and "PRECOND (FRAME_INFER PP (P l t x) Fa)"
      and "PRECOND (mono_body F)"
      and "VC (wf V)"
      and "\<And>D t x. (\<And>t' x. gen_htriple' \<alpha> (\<lambda>l. P l t' x \<and>* \<up>((t', t) \<in> V)) (RCALL D x) (\<lambda>l. X l x) (\<lambda>l. Q l x)) \<Longrightarrow> gen_htriple' \<alpha> (\<lambda>l. P l t x) (F (RCALL D) x) (\<lambda>l. X l x) (\<lambda>l. Q l x)"
      and "\<And>s' r. STATE \<alpha> (Q l x r \<and>* Fa) s' \<Longrightarrow> C r s'"
      and "X l x s \<Longrightarrow> X'"
    shows "gwp (gen_recI \<alpha> V P X Q F x) s X' C"    
    apply (rule sspec_recI_aux[vcg_xformed,OF _ _ _ STATE_FRAME_INFER])
    using assms unfolding RCALL_def
    apply auto
    done
  
  lemma sspec_RCALL[THEN RAWI, vcg_rules]:
    assumes "STATE \<alpha> PP s"
    assumes "PRECOND (ASM (gen_htriple' \<alpha> P (RCALL f x) (\<lambda>l. X l x) (\<lambda>l. Q l x)))"
    assumes "PRECOND (FRAME_INFER PP (P l) F)"
    assumes "\<And>r s'. STATE \<alpha> (Q l x r ** F) s' \<Longrightarrow> C r s'"
    assumes "X l x s \<Longrightarrow> X'"
    shows "gwp (RCALL f x) s X' C"
    apply (rule htriple_sspecI)
    using assms(2) apply (simp add: PRECOND_def; fail)
    apply (rule STATE_FRAME_INFER)
    apply fact+
    done
    
    
  
    
  
subsubsection \<open>While\<close>
(* Pure annotation *)  
definition 
  gen_WhileMI :: "('bi\<Rightarrow>'b\<Rightarrow>bool) \<Rightarrow> 't rel \<Rightarrow> ('t \<Rightarrow> 'a \<Rightarrow> 'bi \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> (bool, 'b) M) \<Rightarrow> ('a \<Rightarrow> ('a, 'b) M) \<Rightarrow> 'a \<Rightarrow> ('a, 'b) M"
  where
  "gen_WhileMI \<alpha> V I b c \<sigma> \<equiv> WhileM 
    (\<lambda>\<sigma>. doN { b \<sigma> }) 
    (\<lambda>\<sigma>. doN { c \<sigma> })
    \<sigma>"

lemma WhileMI_cong: "b=b' \<Longrightarrow> c=c' \<Longrightarrow> \<sigma>=\<sigma>' \<Longrightarrow> gen_WhileMI \<alpha> V I b c \<sigma> = gen_WhileMI \<alpha> V I b' c' \<sigma>'"
  by auto

lemma sspec_WhileMI[vcg_rules]:
  assumes I0: "STATE \<alpha> (I t \<sigma> ** \<Gamma>) s"
  assumes WF: "VC (wf V)"
  assumes S: "\<And>t \<sigma> s. \<lbrakk>STATE \<alpha> (I t \<sigma> ** \<Gamma>) s\<rbrakk> \<Longrightarrow> gwp (b \<sigma>) s X (\<lambda>r s.
      (PURE r \<longrightarrow> gwp (c \<sigma>) s X (\<lambda>\<sigma>' s'. \<exists>t'. POSTCOND \<alpha> (I t' \<sigma>' ** \<up>((t',t)\<in>V) ** \<Gamma>) s'))
    \<and> (PURE (\<not>r) \<longrightarrow> C \<sigma> s)  
    )"
  shows "gwp (gen_WhileMI \<alpha> V I b c \<sigma>) s X C"  
  unfolding gen_WhileMI_def
  apply (rule sspec_WhileM[where I=I, OF I0 WF S])
  .
  
  
(*  
  
oops  
  
xxx, ctd here:
  integrate with memory manager, 
  and do usability tests of VCG (htriple-part)
  
  then explore the two refinement modes!
    in particular, try to do one-sided symbolic execution of bind-chains,
    to do refinement proofs on a basic block level
  


declare_solver STATE_infer:
  STATE_FRAME_INFER
  RSTATE_FRAME_INFER
  = assumption, (frame; fail)
  
\<rightarrow> make the frame-inference a precondition!  
  
  
  
oops  
xxx, ctd here: 
  set up VCG:
    rule preprocessor (srefine/htriple \<rightarrow> srefinec/sspecc)
    main loop, applying rules and invoking solvers
  
   
  
  

print_solvers

oops
  ctd here: 
    integrate frame inference to build VCG
    integrate memory manager
    

  xxx, ctd here: refinement of specification by structured program








    
    
oops
  xxx, ctd here:
    instantiation to sep-logic
    
      
  
oops
  xxx: at this point, we have refinement and Hoare-style rules for a basic return,bind,rec language.

  TODO:
    state-monad also needs get and set! DONE
    
    we need rules for lifting HOL-combinators if, let, case-rules DONE
      \<rightarrow> automation for datatype splitting!
    

    derived combinators DONE + automation for them (?)
          
    instantiate refinement and spec to separation logic
      lift relators to this instantiation
    
    try generic memory model based on HOL+ZF *(afp)!
      allocation and freeing of cells, containing arbitrary small types ( array \<rightarrow> 'a list! )
      pointer is some infinite (small) type
      memory: pointer \<rightarrow> embedded-small-type
        note: it would be nice if we cannot convert between embeddings of different types! 
          (Otherwise, the conversion might not match what actually happens in memory, and
           semantics based on this model may be unsound!)
          \<rightarrow> like imperative HOL, we can use typerep for that!
      \<longrightarrow> search if we have started something like this already!
      term \<open>TYPEREP (int)\<close>
    
oops    
  xxx: structural rule for rec! 
    then: Hoare-rules

    +: How does rf2_mkpw-automation look like for lemmas like run_return?
            
     
  oops  
    
  xxx, ctd here: (lift from lemmas on mres!)
  
    use data refinement to set up program refinement.
    see notes!
    
    try to get structural rules and Hoare-style rules!
      but, ideally, with the same mechanism, i.e., goals of form

      ARG_STATE_RELS (si,s) \<Longrightarrow> rel_mres (RES_STATE_REL) (run mi si) (run m s)
      
      for specification, m will be specm, and ARG_STATE_RELS will be \<subseteq>identity? Does m have state?
      
        
  
  
    recursion
    
    monotonicity of bind, recursion
  
    
    then think about refinement!
    
      Can we, in general, reason about program refinement by 
      
        wp mi (%x. inres x m )
      
    
    
      we have data refinement on mres, seems to be well established!
      
      How does refinement generalize to state?
      \<rightarrow> relation between states. 
         we may need different relations for pre/post state!
         
         something like:
           if pre-states are in relation, 
           each possible concrete result\<times>state is in relation with an abstract result\<times>state
      
      
*)


end
