theory Sep_Generic_Wp
imports 
  "../lib/Sep_Algebra_Add" 
  "../lib/Frame_Infer"
  "../lib/MM/MMonad"
begin


context intf_consistent_loc begin  
  
  lemma intf_r_alloc: "addr.block`r \<subseteq> Collect (is_ALLOC s) \<union> a"
    by (smt (verit, best) Un_def cell.exhaust_disc image_subset_iff intf_ref_not_freed intf_ref_not_fresh is_FRESH'_eq mem_Collect_eq sup_ge1)
  
  lemma intf_w_alloc: "addr.block`w \<subseteq> Collect (is_ALLOC s) \<union> a"
    by (smt (verit, best) Un_def cell.exhaust_disc image_subset_iff intf_ref_not_freed intf_ref_not_fresh is_FRESH'_eq mem_Collect_eq sup_ge1)

end

section \<open>General VCG Setup for Separation Logic\<close>

locale generic_wp =
  fixes wp :: "'c \<Rightarrow> ('r \<Rightarrow> intf \<Rightarrow> 's \<Rightarrow> bool) \<Rightarrow> 's \<Rightarrow> bool"
  assumes wp_comm_inf: "inf (wp c Q) (wp c Q') = wp c (inf Q Q')"
begin

  lemma wp_comm_conj: "wp c (\<lambda>r i. Q r i and Q' r i) s \<longleftrightarrow> wp c Q s \<and> wp c Q' s"
    using wp_comm_inf[of c Q Q']
    unfolding inf_fun_def inf_bool_def by metis

  lemma wp_comm_conjI: "\<lbrakk> wp c Q s; wp c Q' s \<rbrakk> \<Longrightarrow> wp c (\<lambda>r i s. Q r i s \<and> Q' r i s) s"
    using wp_comm_inf[of c Q Q']
    unfolding inf_fun_def inf_bool_def by metis


  lemma wp_mono: "Q\<le>Q' \<Longrightarrow> wp c Q \<le> wp c Q'"
    by (metis (mono_tags) antisym_conv le_inf_iff order_refl wp_comm_inf)

  lemma wp_monoI:
    assumes "wp c Q s"
    assumes "\<And>r i x. Q r i x \<Longrightarrow> Q' r i x"
    shows "wp c Q' s"
    using assms wp_mono[of Q Q' c]
    by (metis le_funI predicate1D predicate1I wp_mono)
    
end

hide_const (open) NEMonad.wp

definition wp where "wp c Q s \<equiv> NEMonad.wp (run c s) (\<lambda>(r,i,s). Q r i s)"

lemma pw_wp[pw_init]: "wp c Q s \<longleftrightarrow> \<not>is_fail (run c s) \<and> (\<forall>r i s'. is_res (run c s) (r,i,s') \<longrightarrow> Q r i s')"
  unfolding wp_def by pw


lemma wp_cons: "\<lbrakk> wp c Q s; \<And>r i s'. Q r i s' \<Longrightarrow> Q' r i s' \<rbrakk> \<Longrightarrow> wp c Q' s"  
  by pw
  
lemma wp_false[simp]: "\<not>wp c (\<lambda>_ _ _. False) s" by pw

interpretation generic_wp wp 
  apply unfold_locales 
  unfolding fun_eq_iff inf_fun_def inf_bool_def
  by pw

lemma wp_return[vcg_normalize_simps]: "wp (Mreturn x) Q s \<longleftrightarrow> Q x 0 s" by pw

lemma wp_fail[vcg_normalize_simps]: "\<not> wp (Mfail) Q s" by pw 

lemma wp_assert[vcg_normalize_simps]: "wp (Massert \<Phi>) Q s \<longleftrightarrow> \<Phi> \<and> Q () 0 s" by pw
  
lemma wp_bind[vcg_normalize_simps]: "wp (doM {x\<leftarrow>m; f x}) Q s = wp m (\<lambda>x i. wp (f x) (\<lambda>x' i'. Q x' (i+i'))) s"  
  by (pw; blast)
  
lemma wp_par:  
  assumes "wp m\<^sub>1 Q\<^sub>1 s"
  assumes "wp m\<^sub>2 Q\<^sub>2 s"
  assumes "\<And>r\<^sub>1 s\<^sub>1 i\<^sub>1 r\<^sub>2 s\<^sub>2 i\<^sub>2. \<lbrakk> 
      intf_consistent s i\<^sub>1 s\<^sub>1; intf_consistent s i\<^sub>2 s\<^sub>2; spar_feasible i\<^sub>1 i\<^sub>2;
      Q\<^sub>1 r\<^sub>1 i\<^sub>1 s\<^sub>1; Q\<^sub>2 r\<^sub>2 i\<^sub>2 s\<^sub>2 
    \<rbrakk> 
    \<Longrightarrow> intf_norace i\<^sub>1 i\<^sub>2 \<and> Q (r\<^sub>1,r\<^sub>2) (i\<^sub>1+i\<^sub>2) (combine_states s\<^sub>1 i\<^sub>2 s\<^sub>2)"
  shows "wp (Mpar m\<^sub>1 m\<^sub>2) Q s"
  using assms apply -
  apply pw 
  apply (meson res_run_consistentI)
  apply (meson res_run_consistentI)
  done
  
lemma wp_malloc[vcg_normalize_simps]: "wp (Mmalloc xs) Q s \<longleftrightarrow> (\<forall>r. is_FRESH s r \<longrightarrow> Q r (intf_a r) (addr_alloc xs r s))"  
  supply [pw_simp] = malloc_def
  by pw
  
lemma wp_free[vcg_normalize_simps]: "wp (Mfree b) Q s \<longleftrightarrow> is_ALLOC s b \<and> Q () (intf_f b) (addr_free b s)"  
  supply [pw_simp] = mfree_def
  by pw
  
lemma wp_load[vcg_normalize_simps]: "wp (Mload a) Q s \<longleftrightarrow> is_valid_addr s a \<and> Q (get_addr s a) (intf_r a) s"  
  supply [pw_simp] = mload_def
  by pw

lemma wp_mstore[vcg_normalize_simps]: "wp (Mstore a x) Q s \<longleftrightarrow> is_valid_addr s a \<and> Q () (intf_w a) (put_addr s a x)"  
  supply [pw_simp] = mstore_def
  by pw

lemma wp_valid_addr[vcg_normalize_simps]: "wp (Mvalid_addr a) Q s \<longleftrightarrow> is_valid_addr s a \<and> Q () (intf_r a) s"  
  supply [pw_simp] = mvalid_addr_def
  by pw
  
(*lemma wp_valid_ptr_addr[vcg_normalize_simps]: "wp (Mvalid_ptr_addr a) Q s \<longleftrightarrow> is_valid_ptr_addr s a \<and> Q () (intf_e (addr.block a)) s"  
  supply [pw_simp] = mvalid_ptr_addr_def
  by pw
*)    

definition "mem_eq_on A B s\<^sub>1 s\<^sub>2 \<equiv>
  (\<forall>b\<in>B. is_ALLOC s\<^sub>1 b = is_ALLOC s\<^sub>2 b \<and> block_size s\<^sub>1 b = block_size s\<^sub>2 b)
\<and> (\<forall>a\<in>A. is_valid_addr s\<^sub>1 a = is_valid_addr s\<^sub>2 a \<and> get_addr s\<^sub>1 a = get_addr s\<^sub>2 a)"

lemma mem_eq_on_refl[simp]: "mem_eq_on A B s s"
  by (auto simp: mem_eq_on_def)

lemma mem_eq_on_trans[trans]: "mem_eq_on A B s\<^sub>1 s\<^sub>2 \<Longrightarrow> mem_eq_on A B s\<^sub>2 s\<^sub>3 \<Longrightarrow> mem_eq_on A B s\<^sub>1 s\<^sub>3"
  by (auto simp: mem_eq_on_def)
  
lemma mem_eq_on_mono: "A\<subseteq>A' \<Longrightarrow> B\<subseteq>B' \<Longrightarrow> mem_eq_on A' B' s\<^sub>1 s\<^sub>2 \<Longrightarrow> mem_eq_on A B s\<^sub>1 s\<^sub>2"  
  by (auto simp: mem_eq_on_def)
  
locale abs_sep =  
  fixes \<alpha> :: "'v memory \<Rightarrow> 'a::{stronger_sep_algebra}"
    and addrs :: "'a \<Rightarrow> addr set"    (* Addresses *)
    and blocks :: "'a \<Rightarrow> block set"  (* Blocks *)
  assumes addr_add[simp]: "a##b \<Longrightarrow> addrs (a+b) = addrs a \<union> addrs b"
      and addrs_\<alpha>: "addrs (\<alpha> s) = Collect (is_valid_addr s)"
  assumes block_add[simp]: "a##b \<Longrightarrow> blocks (a+b) = blocks a \<union> blocks b"
      and blocks_\<alpha>: "blocks (\<alpha> s) = Collect (is_ALLOC s)"
  assumes dj_iff: "a##b \<longleftrightarrow> addrs a \<inter> addrs b = {} \<and> blocks a \<inter> blocks b = {}"    

  assumes zero_char: "addrs as = {} \<Longrightarrow> blocks as = {} \<Longrightarrow> as=0"
  
  assumes xfer_rule: "\<lbrakk> as ## f; \<alpha> s = as + f; mem_eq_on (addrs as) (blocks as) s s' \<rbrakk> 
    \<Longrightarrow> \<exists>f'. as ## f' \<and> \<alpha> s' = as + f'"

  assumes split_rule: "\<lbrakk> a##b; a##f\<^sub>1; b##f\<^sub>2; a+f\<^sub>1 = b+f\<^sub>2 \<rbrakk> \<Longrightarrow> \<exists>f'. b ## f' \<and> f\<^sub>1 = b + f'"  
          
begin


  lemma split_rule':
    fixes a :: 'a
    assumes "a##b" "a##f\<^sub>1" "b##f\<^sub>2"
    assumes "a+f\<^sub>1 = b+f\<^sub>2"
    obtains f' where "b ## f'" "f\<^sub>1 = b + f'" 
    using split_rule[OF assms] by blast


  abbreviation "mem_eq_on' as \<equiv> mem_eq_on (addrs as) (blocks as)"

  lemma mem_eq_xfer_pair:
    assumes "as\<^sub>1 ## f\<^sub>1" "as\<^sub>2 ## f\<^sub>2" "as\<^sub>1 ## as\<^sub>2"
    assumes A1: "\<alpha> s\<^sub>1 = as\<^sub>1 + f\<^sub>1"
    assumes A2: "\<alpha> s\<^sub>2 = as\<^sub>2 + f\<^sub>2"
    assumes "Collect (is_valid_addr s') = addrs as\<^sub>1 \<union> addrs as\<^sub>2"
    assumes "Collect (is_ALLOC s') = blocks as\<^sub>1 \<union> blocks as\<^sub>2"
    
    assumes E1: "mem_eq_on' as\<^sub>1 s\<^sub>1 s'" 
    assumes E2: "mem_eq_on' as\<^sub>2 s\<^sub>2 s'"
    
    shows "\<alpha> s' = as\<^sub>1 + as\<^sub>2"
  proof -
    from xfer_rule[OF \<open>as\<^sub>1 ## f\<^sub>1\<close> A1 E1] obtain f\<^sub>1' where "as\<^sub>1 ## f\<^sub>1'" "\<alpha> s' = as\<^sub>1 + f\<^sub>1'" by blast
    from xfer_rule[OF \<open>as\<^sub>2 ## f\<^sub>2\<close> A2 E2] obtain f\<^sub>2' where "as\<^sub>2 ## f\<^sub>2'" "\<alpha> s' = as\<^sub>2 + f\<^sub>2'" by blast
    
    obtain f' where "as\<^sub>2 ## f'" "f\<^sub>1' = as\<^sub>2 + f'" 
      apply (rule split_rule'[of as\<^sub>1 as\<^sub>2 f\<^sub>1' f\<^sub>2'])
      apply (rule assms \<open>as\<^sub>1 ## f\<^sub>1'\<close> \<open>as\<^sub>2 ## f\<^sub>2'\<close>)+
      using \<open>\<alpha> s' = as\<^sub>1 + f\<^sub>1'\<close> \<open>\<alpha> s' = as\<^sub>2 + f\<^sub>2'\<close> apply auto[1]
      by blast
    
    hence AS': "\<alpha> s' = as\<^sub>1 + as\<^sub>2 + f'" using \<open>\<alpha> s' = as\<^sub>1 + f\<^sub>1'\<close> 
      by (metis \<open>as\<^sub>1 ## f\<^sub>1'\<close> sep_add_assoc sep_add_disj_eq)
    
    hence "addrs f' = {}"
      by (metis Un_Int_eq(4) \<open>as\<^sub>1 ## f\<^sub>1'\<close> \<open>as\<^sub>2 ## f'\<close> \<open>f\<^sub>1' = as\<^sub>2 + f'\<close> addr_add addrs_\<alpha> assms(6) dj_iff sep_disj_add_eq sep_disj_commuteI)
      
    moreover have "blocks f' = {}"
      by (metis Un_Int_eq(4) \<open>\<alpha> s' = as\<^sub>1 + as\<^sub>2 + f'\<close> 
        \<open>as\<^sub>1 ## f\<^sub>1'\<close> \<open>as\<^sub>2 ## f'\<close> \<open>f\<^sub>1' = as\<^sub>2 + f'\<close> blocks_\<alpha> 
         assms(3) assms(7) block_add dj_iff sep_add_commute 
         sep_disj_addI2 sep_disj_commuteI)
  
    ultimately have "f'=0" by (rule zero_char)
    with AS' show ?thesis by simp
  qed
  
  lemma addr_dj: "a##b \<Longrightarrow> addrs a \<inter> addrs b = {}" by (auto simp: dj_iff) 

  lemma block_dj: "a##b \<Longrightarrow> blocks a \<inter> blocks b = {}" by (auto simp: dj_iff)
  

  lemma addrsZ[simp]: "addrs 0 = {}"
    using addr_dj disjoint_zero_sym by blast

  lemma addr_djI[intro]:
    "a##b \<Longrightarrow> x\<in>addrs a \<Longrightarrow> x\<notin>addrs b"  
    "a##b \<Longrightarrow> x\<in>addrs b \<Longrightarrow> x\<notin>addrs a"  
    using addr_dj by auto
    

      
text \<open>Eliminating the interference, only maintaining a set of forbidden addresses\<close>  

definition "STATE asf P s \<equiv> \<exists>as. as ## asf \<and> \<alpha> s = as+asf \<and> P as"
    
definition wpa where [pw_init]: "wpa asf c Q s \<equiv> wp c (\<lambda>r i s'. 
    Q r s' 
  \<and> intf.r i \<union> intf.w i \<subseteq> -addrs asf
  \<and> intf.f i \<subseteq> - (addr.block`addrs asf \<union> blocks asf)
) s"

lemma wpa_monoI:
  assumes "wpa A c Q s"
  assumes "\<And>r s'. \<lbrakk> Q r s' \<rbrakk> \<Longrightarrow> Q' r s'"
  assumes "blocks A' \<subseteq> blocks A" "addrs A' \<subseteq> addrs A"
  shows "wpa A' c Q' s"
  using assms unfolding wpa_def
  by (blast intro: wp_monoI)



definition "htriple P c Q \<equiv> (\<forall>s asf. STATE asf P s 
  \<longrightarrow> wpa asf c (\<lambda>r s'. STATE asf (Q r) s') s)"

lemma htripleI[intro?]:
  assumes "\<And>s asf. STATE asf P s \<Longrightarrow> wpa asf c (\<lambda>r s'. STATE asf (Q r) s') s"
  shows "htriple P c Q"
  using assms unfolding htriple_def by blast

lemma htripleD: 
  assumes "htriple P c Q"
  assumes "STATE asf P s"
  shows "wpa asf c (\<lambda>r s'. STATE asf (Q r) s') s"
  using assms unfolding htriple_def by blast


lemma wpa_false[vcg_normalize_simps]: "\<not>wpa A m (\<lambda>_ _. False) s"  
  by (simp add: wpa_def vcg_normalize_simps)

lemma wpa_return[vcg_normalize_simps]: "wpa asf (Mreturn x) Q s \<longleftrightarrow> Q x s" by pw

lemma wpa_fail[vcg_normalize_simps]: "\<not> wpa asf (Mfail) Q s" by pw 

lemma wpa_assert[vcg_normalize_simps]: "wpa asf (Massert \<Phi>) Q s \<longleftrightarrow> \<Phi> \<and> Q () s" by pw

lemma wpa_bindI[vcg_decomp_rules]: "wpa asf m (\<lambda>x. wpa asf (f x) Q) s \<Longrightarrow> wpa asf (doM {x\<leftarrow>m; f x}) Q s"  
  unfolding wpa_def
  apply (simp add: vcg_normalize_simps)
  apply (erule wp_cons)
  apply clarify
  apply (erule wp_cons)
  apply auto
  done
  
lemma wpa_ifI[vcg_decomp_rules]: 
  assumes "b \<Longrightarrow> wpa asf c Q s"
  assumes "\<not>b \<Longrightarrow> wpa asf e Q s"
  shows "wpa asf (if b then c else e) Q s"  
  using assms by auto
  
  
(* TODO: Move *)  
declare sep_disj_commuteI[sym]  
  

  lemma STATE_triv[simp]: "\<not>STATE asf sep_false s"
    by (simp add: STATE_def)
    
  lemma STATE_monoI: "STATE asf P s \<Longrightarrow> P \<turnstile> P' \<Longrightarrow> STATE asf P' s"  
    unfolding STATE_def entails_def by blast
    
  lemma STATE_frame1:
    "STATE asf (P**F) s \<Longrightarrow> \<exists>asf'. asf'##asf \<and> F asf' \<and> STATE (asf'+asf) P s"  
    unfolding STATE_def
    apply (auto simp: sep_algebra_simps sep_conj_def)
    by fastforce

  lemma STATE_frame2: 
    assumes "STATE (asf'+asf) P s"
    assumes "asf' ## asf"  
    assumes "F asf'"
    shows "STATE asf (P**F) s"
    using assms
    unfolding STATE_def
    apply (auto simp: sep_algebra_simps sep_conj_def) 
    by (metis assms(2) sep_add_assoc sep_add_commute sep_disj_add_eq)

  lemma htriple_triv[simp, intro!]: "htriple sep_false c Q"
    by (auto simp: htriple_def)
      
  lemma cons_rule:
    assumes "htriple P c Q"
    assumes "\<And>s. P' s \<Longrightarrow> P s"
    assumes "\<And>r s. Q r s \<Longrightarrow> Q' r s"
    shows "htriple P' c Q'"
  proof
    fix asf s
    assume "STATE asf P' s"
    with STATE_monoI assms(2) have "STATE asf P s" unfolding entails_def by blast
    from htripleD[OF assms(1) this] have "wpa asf c (\<lambda>r. STATE asf (Q r)) s" .
    then show "wpa asf c (\<lambda>r. STATE asf (Q' r)) s"
      apply (rule wpa_monoI)
      apply (erule STATE_monoI)
      using assms(3) unfolding entails_def 
      by auto
      
  qed  
      
  lemma cons_post_rule:
    assumes "htriple P c Q"
    assumes "\<And>r s. Q r s \<Longrightarrow> Q' r s"
    shows "htriple P c Q'"
    using cons_rule assms by blast
    
  lemma apply_htriple_rule:
    assumes "STATE asf (P**F) s"
    assumes "htriple P c Q"
    assumes "\<And>s' r. STATE asf (Q r ** F) s' \<Longrightarrow> Q' r s'"
    shows "wpa asf c Q' s"  
  proof -
    from STATE_frame1[OF assms(1)] obtain asf' where 
      D: "asf' ## asf" "F asf'" 
      and S: "STATE (asf' + asf) P s" 
      by blast
  
    from htripleD[OF assms(2) S] have "wpa (asf'+asf) c (\<lambda>r s'. STATE (asf' + asf) (Q r) s') s" .
    moreover have "\<lbrakk>STATE (asf' + asf) (Q r) s'\<rbrakk> \<Longrightarrow> STATE asf (Q r ** F) s'" for r s'
      by (rule STATE_frame2; (assumption|rule D)?)
    ultimately have "wpa (asf)  c (\<lambda>r s'. STATE asf (Q r ** F) s') s"
      apply (rule wpa_monoI) by (auto simp: D)
    then show ?thesis using assms(3)
      by (rule wpa_monoI) auto
  qed

        
  lemma frame_rule: "htriple P c Q \<Longrightarrow> htriple (P ** F) c (\<lambda>r. Q r ** F)"
    apply rule
    apply (erule apply_htriple_rule)
    .







lemma ht_par:
  assumes HT1: "htriple P\<^sub>1 m\<^sub>1 Q\<^sub>1"
  assumes HT2: "htriple P\<^sub>2 m\<^sub>2 Q\<^sub>2"
  shows "htriple (P\<^sub>1**P\<^sub>2) (m\<^sub>1 \<parallel>\<^sub>M m\<^sub>2) (\<lambda>(r\<^sub>1,r\<^sub>2). Q\<^sub>1 r\<^sub>1 ** Q\<^sub>2 r\<^sub>2)"
proof
  fix s asf
  
  assume "STATE asf (P\<^sub>1 \<and>* P\<^sub>2) s"
  then obtain as\<^sub>1 as\<^sub>2 where 
        [simp]: "as\<^sub>1 ## as\<^sub>2" "(as\<^sub>1 + as\<^sub>2) ## asf" 
    and S_SPLIT: "\<alpha> s = as\<^sub>1 + as\<^sub>2 + asf" 
    and PRECOND[simp]: "P\<^sub>1 as\<^sub>1" "P\<^sub>2 as\<^sub>2"
    unfolding STATE_def by (auto simp: sep_conj_def)

  have [simp, symmetric, simp]: "as\<^sub>1 ## asf" "as\<^sub>2 ## asf"
    using \<open>as\<^sub>1 ## as\<^sub>2\<close> \<open>as\<^sub>1 + as\<^sub>2 ## asf\<close> sep_add_disjD apply blast
    using \<open>as\<^sub>1 ## as\<^sub>2\<close> \<open>as\<^sub>1 + as\<^sub>2 ## asf\<close> sep_add_disjD apply blast
    done   
      
    
  have S1: "STATE (as\<^sub>2 + asf) P\<^sub>1 s"  
    unfolding STATE_def by (auto simp: sep_conj_def sep_algebra_simps S_SPLIT intro!: exI[where x="as\<^sub>1"])
    
  have S2: "STATE (as\<^sub>1 + asf) P\<^sub>2 s"  
    unfolding STATE_def by (auto simp: sep_conj_def sep_algebra_simps S_SPLIT intro!: exI[where x="as\<^sub>2"])
    

  have ADDRS: "addrs (\<alpha> s) = addrs as\<^sub>1 \<union> addrs as\<^sub>2 \<union> addrs asf"
    and ADDRS_DJ: "addrs as\<^sub>1 \<inter> (addrs as\<^sub>2 \<union> addrs asf) = {}" "addrs as\<^sub>2 \<inter> addrs asf = {}"
    by (simp_all add: Int_Un_distrib addr_dj S_SPLIT)
    
  have BLOCKS: "blocks (\<alpha> s) = blocks as\<^sub>1 \<union> blocks as\<^sub>2 \<union> blocks asf"
    and BLOCKS_DJ: "blocks as\<^sub>1 \<inter> (blocks as\<^sub>2 \<union> blocks asf) = {}" "blocks as\<^sub>2 \<inter> blocks asf = {}"
    by (simp_all add: Int_Un_distrib block_dj S_SPLIT)
    
    
  note WPA1 = htripleD[OF HT1 S1]
  note WPA2 = htripleD[OF HT2 S2]

  
  (* Consistency properties *)
  

  
  show "wpa asf (m\<^sub>1 \<parallel>\<^sub>M m\<^sub>2) (\<lambda>r. STATE asf (case r of (r\<^sub>1, r\<^sub>2) \<Rightarrow> Q\<^sub>1 r\<^sub>1 \<and>* Q\<^sub>2 r\<^sub>2)) s"
    using WPA1 WPA2
    apply (simp add: wpa_def)
    apply (erule (1) wp_par, thin_tac "wp _ _ _", clarsimp)
    proof goal_cases
      case (1 x\<^sub>1 s\<^sub>1 i\<^sub>1 x\<^sub>2 s\<^sub>2 i\<^sub>2)
    
      interpret feasible_parallel_execution s s\<^sub>1 i\<^sub>1 s\<^sub>2 i\<^sub>2 using 1 apply unfold_locales by simp_all

      have [simp]: "valid_addrs_s = addrs as\<^sub>1 \<union> addrs as\<^sub>2 \<union> addrs asf"
        using ADDRS addrs_\<alpha> by auto
            
      have [simp]: "valid_blocks_s = blocks as\<^sub>1 \<union> blocks as\<^sub>2 \<union> blocks asf"  
        using BLOCKS blocks_\<alpha> by auto
        
      show ?case proof (intro conjI)
      
      
        show 
          "r\<^sub>1 \<subseteq> - addrs asf"
          "r\<^sub>2 \<subseteq> - addrs asf"
          "w\<^sub>1 \<subseteq> - addrs asf"
          "w\<^sub>2 \<subseteq> - addrs asf"
          "f\<^sub>1 \<subseteq> - addr.block ` addrs asf - blocks asf"
          "f\<^sub>2 \<subseteq> - addr.block ` addrs asf - blocks asf" 
          using 1 
          by auto

        have A1_DJ: "c1.allocated_addrs \<inter> (addrs as\<^sub>1 \<union> addrs as\<^sub>2 \<union> addrs asf) = {}"  
          using c1.a_dj_valid by auto
          
        have A2_DJ: "c2.allocated_addrs \<inter> (addrs as\<^sub>1 \<union> addrs as\<^sub>2 \<union> addrs asf) = {}"  
          using c2.a_dj_valid by auto
          
          
        show "intf_norace i\<^sub>1 i\<^sub>2" 
        proof -
                  
          have RW1_SS: "r\<^sub>1\<union>w\<^sub>1 \<subseteq> addrs as\<^sub>1 \<union> c1.allocated_addrs" 
            using c1.rw_valid_or_alloc
            using \<open>w\<^sub>1 \<subseteq> - addrs as\<^sub>2 - addrs asf\<close> \<open>r\<^sub>1 \<subseteq> - addrs as\<^sub>2 - addrs asf\<close>
            by auto
          
          have RW2_SS: "r\<^sub>2\<union>w\<^sub>2 \<subseteq> addrs as\<^sub>2 \<union> {a. addr.block a \<in> a\<^sub>2}"  
            using c2.rw_valid_or_alloc
            using \<open>w\<^sub>2 \<subseteq> - addrs as\<^sub>1 - addrs asf\<close> \<open>r\<^sub>2 \<subseteq> - addrs as\<^sub>1 - addrs asf\<close> 
            by auto
  
          have G1: "(r\<^sub>1\<union>w\<^sub>1) \<inter> (r\<^sub>2\<union>w\<^sub>2) = {}" 
            using RW1_SS RW2_SS A1_DJ A2_DJ alloc_disj 
            using \<open>r\<^sub>1 \<subseteq> - addrs as\<^sub>2 - addrs asf\<close> \<open>r\<^sub>2 \<subseteq> - addrs as\<^sub>1 - addrs asf\<close>
            using \<open>w\<^sub>1 \<subseteq> - addrs as\<^sub>2 - addrs asf\<close> \<open>w\<^sub>2 \<subseteq> - addrs as\<^sub>1 - addrs asf\<close>
            by blast
            
          have F1_SS: "f\<^sub>1 \<subseteq> blocks as\<^sub>1 \<union> a\<^sub>1" 
            using c1.f_valid_or_alloc  \<open>f\<^sub>1 \<subseteq> - blocks as\<^sub>2 - blocks asf\<close>
            by auto
            
          have F2_SS: "f\<^sub>2 \<subseteq> blocks as\<^sub>2 \<union> a\<^sub>2" 
            using c2.f_valid_or_alloc  \<open>f\<^sub>2 \<subseteq> - blocks as\<^sub>1 - blocks asf\<close>
            by auto
            
          have A1_BDJ: "a\<^sub>1 \<inter> blocks as\<^sub>2 = {}"  
            using c1.a_dj_alloc by fastforce
            
          have A2_BDJ: "a\<^sub>2 \<inter> blocks as\<^sub>1 = {}"  
            using c2.a_dj_alloc by fastforce
            
          have G2: "f\<^sub>1\<inter>f\<^sub>2={}" 
            using \<open>f\<^sub>1 \<subseteq> - blocks as\<^sub>2 - blocks asf\<close> A2_BDJ alloc_disj F1_SS F2_SS by blast
            
          have G3: "addr.block`(r\<^sub>1\<union>w\<^sub>1) \<inter> f\<^sub>2 = {}"
            using RW1_SS A1_BDJ alloc_disj F2_SS \<open>f\<^sub>2 \<subseteq> - addr.block ` (addrs as\<^sub>1 \<union> addrs asf)\<close>
            by blast
            
          have G4: "addr.block`(r\<^sub>2\<union>w\<^sub>2) \<inter> f\<^sub>1 = {}"
            using RW2_SS A2_BDJ alloc_disj F1_SS \<open>f\<^sub>1 \<subseteq> - addr.block ` (addrs as\<^sub>2 \<union> addrs asf)\<close>
            by blast
                    
          show "intf_norace i\<^sub>1 i\<^sub>2"
            unfolding intf_norace_def Let_def
            apply simp
            using G1 G2 G3 G4
            by blast
        qed                  

        then interpret valid_parallel_execution s s\<^sub>1 i\<^sub>1 s\<^sub>2 i\<^sub>2 using 1 apply unfold_locales by simp_all
        
        show "STATE asf (Q\<^sub>1 x\<^sub>1 \<and>* Q\<^sub>2 x\<^sub>2) s'"
        proof -
          note \<open>STATE (as\<^sub>2 + asf) (Q\<^sub>1 x\<^sub>1) s\<^sub>1\<close> \<open>STATE (as\<^sub>1 + asf) (Q\<^sub>2 x\<^sub>2) s\<^sub>2\<close>
          then obtain as\<^sub>1' as\<^sub>2' where 
                A12_FMT:
            "\<alpha> s\<^sub>1 = as\<^sub>1' + as\<^sub>2 + asf"
            "\<alpha> s\<^sub>2 = as\<^sub>1 + as\<^sub>2' + asf"
            and ASx'_DJS:
            "as\<^sub>1' ## (as\<^sub>2 + asf)"
            "as\<^sub>2' ## (as\<^sub>1 + asf)"
            and 
            "Q\<^sub>1 x\<^sub>1 as\<^sub>1'"
            "Q\<^sub>2 x\<^sub>2 as\<^sub>2'"
            unfolding STATE_def
            by (auto simp: sep_algebra_simps )
            
          from ASx'_DJS have DJ'[simp]: 
            "as\<^sub>1' ## as\<^sub>2" "as\<^sub>2' ## as\<^sub>1" "as\<^sub>1' ## asf" "as\<^sub>2' ## asf"
            by auto
            
          have [simp]: "as\<^sub>1 ## as\<^sub>2'"
            by (simp add: sep_disj_commuteI)

          have [simp]: 
            "Collect (is_valid_addr s\<^sub>1) = addrs as\<^sub>1' \<union> addrs as\<^sub>2 \<union> addrs asf"
            "Collect (is_valid_addr s\<^sub>2) = addrs as\<^sub>1 \<union> addrs as\<^sub>2' \<union> addrs asf"
            "is_ALLOC s\<^sub>1 b \<longleftrightarrow> b \<in> blocks as\<^sub>1' \<union> blocks as\<^sub>2 \<union> blocks asf"
            "is_ALLOC s\<^sub>2 b \<longleftrightarrow> b \<in> blocks as\<^sub>1 \<union> blocks as\<^sub>2' \<union> blocks asf"
            for b
            using blocks_\<alpha>[of s\<^sub>1] blocks_\<alpha>[of s\<^sub>2]
            by (auto simp flip: addrs_\<alpha> simp: A12_FMT)
        
                        
          have [simp]: "as\<^sub>1' ## as\<^sub>2'" proof -
          
            have ADDRS_DJ':
              "addrs as\<^sub>1' \<inter> (addrs as\<^sub>2 \<union> addrs asf) = {}" "addrs as\<^sub>2 \<inter> addrs asf = {}"  
              using DJ'[unfolded dj_iff] ADDRS_DJ(2)
              by auto
              
            have BLOCKS_DJ':
              "blocks as\<^sub>1' \<inter> (blocks as\<^sub>2 \<union> blocks asf) = {}" "blocks as\<^sub>2 \<inter> blocks asf = {}"  
              using DJ'[unfolded dj_iff] BLOCKS_DJ(2)
              by auto
            
            
            have AS1'_SS: "addrs as\<^sub>1' \<subseteq> c1.allocated_addrs \<union> addrs as\<^sub>1"
              using ADDRS_DJ'(1) c1.valid_addrs'_subset by auto
              
            have AS2'_SS: "addrs as\<^sub>2' \<subseteq> c2.allocated_addrs \<union> addrs as\<^sub>2"
              using c2.valid_addrs'_subset 
              apply auto 
              by (smt (verit, ccfv_threshold) DJ'(4) UnE \<open>as\<^sub>1 ## as\<^sub>2'\<close> addr_djI(2) mem_Collect_eq subset_iff)

            have G1: "addrs as\<^sub>1' \<inter> addrs as\<^sub>2' = {}"
              using alloc_addrs_disj AS1'_SS AS2'_SS A1_DJ A2_DJ ADDRS_DJ'(1)
              by blast


            have BS1'_SS: "blocks as\<^sub>1' \<subseteq> a\<^sub>1 \<union> blocks as\<^sub>1"
              using c1.is_ALLOC'_eq
              apply auto
              by (metis (mono_tags, lifting) BLOCKS BLOCKS_DJ'(1) UnE Un_assoc blocks_\<alpha> 
                    disjoint_iff_not_equal mem_Collect_eq)
              
            have BS2'_SS: "blocks as\<^sub>2' \<subseteq> a\<^sub>2 \<union> blocks as\<^sub>2"
              using c2.is_ALLOC'_eq 
              apply auto 
              by (smt (z3) DJ'(4) UnE \<open>as\<^sub>1 ## as\<^sub>2'\<close> \<open>valid_blocks_s = blocks as\<^sub>1 \<union> blocks as\<^sub>2 \<union> blocks asf\<close> block_dj disjoint_iff mem_Collect_eq)
                        
            have G2: "blocks as\<^sub>1' \<inter> blocks as\<^sub>2' = {}"
              using BS1'_SS BS2'_SS alloc_disj block_dj[OF DJ'(2)] BLOCKS_DJ'(1) by blast
              
            from G1 G2 show ?thesis by (simp add: dj_iff)
          qed

          have AS'_FMT: \<open>\<alpha> s' = as\<^sub>1' + as\<^sub>2' + asf\<close> 
          proof -
          
          
            have MEQ1: "mem_eq_on' (as\<^sub>1' + asf) s\<^sub>1 s'" sorry
            have MEQ2: "mem_eq_on' (as\<^sub>2') s\<^sub>2 s'" sorry
          
            have AS1_EQ: "\<alpha> s\<^sub>1 = as\<^sub>1'+asf + as\<^sub>2" 
              using A12_FMT by (simp add: A12_FMT sep_algebra_simps)

            have AS2_EQ: "\<alpha> s\<^sub>2 = as\<^sub>2'+(asf + as\<^sub>1)" 
              using A12_FMT by (simp add: A12_FMT sep_algebra_simps)
              
            show ?thesis 
              apply (subst mem_eq_xfer_pair[OF _ _ _ AS1_EQ AS2_EQ _ _ MEQ1 MEQ2])
            proof -  
              show "Collect (is_valid_addr s') = addrs (as\<^sub>1' + asf) \<union> addrs as\<^sub>2'"
                apply (clarsimp) sorry
              
              show "{b. is_ALLOC s' b} = blocks (as\<^sub>1' + asf) \<union> blocks as\<^sub>2'" sorry
            
            qed (auto simp: sep_algebra_simps)
          
          qed
          
          have Q12: "(Q\<^sub>1 x\<^sub>1 ** Q\<^sub>2 x\<^sub>2) (as\<^sub>1' + as\<^sub>2')"
            by (meson \<open>Q\<^sub>1 x\<^sub>1 as\<^sub>1'\<close> \<open>Q\<^sub>2 x\<^sub>2 as\<^sub>2'\<close> \<open>as\<^sub>1' ## as\<^sub>2'\<close> sep_conjI)
          
          show ?thesis
            unfolding STATE_def
            apply (rule exI[where x = "as\<^sub>1' + as\<^sub>2'"])
            apply (simp add: AS'_FMT Q12)
            done
            
        qed
      qed
    qed
  qed  
        
  
  
  
subsection \<open>Realizable abstract Predicates\<close>    
  definition "realizable P \<equiv> \<exists>s as asf. P as \<and> \<alpha> s = as + asf \<and> as ## asf"
  
  lemma realizable_conjD: "realizable (P**Q) \<Longrightarrow> realizable P \<and> realizable Q"
    apply (clarsimp simp: realizable_def sep_conj_def sep_algebra_simps sep_conj_c; intro conjI)
    subgoal for s x asf y
      apply (rule exI[where x=s])
      apply (rule exI[where x="x"])
      apply simp
      apply (rule exI[where x="asf+y"])
      by simp
    
    subgoal for s x asf y
      apply (rule exI[where x=s])
      apply (rule exI[where x="y"])
      apply simp
      apply (rule exI[where x="asf+x"])
      by (simp add: sep_algebra_simps)
      
    done

  lemma realizable_emp[simp]: "realizable \<box>"    
    by (auto simp add: realizable_def sep_algebra_simps)
          
  lemma realizable_pure[simp]: "realizable (\<up>\<phi>) \<longleftrightarrow> \<phi>"  
    by (auto simp add: realizable_def sep_algebra_simps)
      
  lemma realizable_extract_pure[simp]: "realizable (\<up>\<phi> ** P) \<longleftrightarrow> \<phi> \<and> realizable P"  
    by (auto simp add: realizable_def sep_algebra_simps)
      
  lemma htriple_false: "htriple P c (\<lambda>r s. False) \<longleftrightarrow> \<not>realizable P"
    unfolding htriple_def STATE_def
    by (auto simp: wpa_false realizable_def)
    
    
  lemma htriple_realizable_preI: 
    assumes "realizable P \<Longrightarrow> htriple P c Q"
    shows "htriple P c Q"
    using assms
    using cons_post_rule htriple_false by fastforce
    
  lemma STATE_realizableI: "STATE f P s \<Longrightarrow> realizable P"  
    unfolding STATE_def realizable_def by auto
    
    
subsection \<open>VCG Setup\<close>    
    
lemma STATE_assn_cong[fri_extract_congs]: "P\<equiv>P' \<Longrightarrow> STATE f P s \<equiv> STATE f P' s" by simp
  
lemma STATE_extract[vcg_normalize_simps]:
  "STATE asf (\<up>\<Phi>) h \<longleftrightarrow> \<Phi> \<and> STATE asf \<box> h"
  "STATE asf (\<up>\<Phi> ** P) h \<longleftrightarrow> \<Phi> \<and> STATE asf P h"
  "STATE asf (EXS x. Q x) h \<longleftrightarrow> (\<exists>x. STATE asf (Q x) h)"
  "STATE asf (\<lambda>_. False) h \<longleftrightarrow> False"
  "STATE asf ((\<lambda>_. False) ** P) h \<longleftrightarrow> False"
  by (auto simp: STATE_def sep_algebra_simps)
 

definition POSTCOND where [vcg_tag_defs]: "POSTCOND \<equiv> STATE"
  
lemma POSTCONDI:
  "STATE asf Q h \<Longrightarrow> POSTCOND asf Q h"
  by (auto simp add: POSTCOND_def)
lemma POSTCOND_cong[vcg_normalize_congs]: "POSTCOND asf Q = POSTCOND asf Q" ..

lemma POSTCOND_triv_simps[vcg_normalize_simps]:
  (* "POSTCOND A asf sep_true h" *)
  "\<not>POSTCOND asf sep_false h"
  unfolding POSTCOND_def STATE_def by auto
  
lemma start_entailsE:
  assumes "STATE asf P h"
  assumes "ENTAILS P P'"
  shows "STATE asf P' h"
  using assms unfolding STATE_def ENTAILS_def entails_def
  by auto

declaration \<open>fn phi =>
   (Basic_VCG.add_xformer (map (Morphism.thm phi) @{thms POSTCONDI},@{binding postcond_xformer},
    fn ctxt => eresolve_tac ctxt (map (Morphism.thm phi) @{thms start_entailsE})
  ))
\<close>

named_theorems htriple_vcg_intros
named_theorems htriple_vcg_intro_congs
definition [vcg_tag_defs]: "DECOMP_HTRIPLE \<phi> \<equiv> \<phi>"
lemma DECOMP_HTRIPLEI: "\<phi> \<Longrightarrow> DECOMP_HTRIPLE \<phi>" unfolding vcg_tag_defs by simp

  lemma htriple_vcg_frame_erule[vcg_frame_erules]:
    assumes S: "STATE asf P' s"
    assumes F: "PRECOND (FRAME P' P F)"
    assumes HT: "htriple P c Q"  
    assumes P: "\<And>r s. STATE asf (Q r ** F) s \<Longrightarrow> PRECOND (EXTRACT (Q' r s))"
    shows "wpa asf c Q' s"
  proof -
    from S F have S': "STATE asf (P**F) s"
      unfolding PRECOND_def FRAME_def
      using STATE_monoI by blast

    show ?thesis
      apply (rule apply_htriple_rule[OF S' HT])
      using P unfolding vcg_tag_defs .
      
  qed    
  
  lemma htriple_vcgI': 
    assumes "\<And>asf s. STATE asf P s \<Longrightarrow> wpa asf c (\<lambda>r. POSTCOND asf (Q r)) s"
    shows "htriple P c Q"
    apply (rule htripleI)
    using assms unfolding vcg_tag_defs .
  
  lemma htriple_vcgI[htriple_vcg_intros]: 
    assumes "\<And>asf s. STATE asf P s \<Longrightarrow> EXTRACT (wpa asf c (\<lambda>r. POSTCOND asf (Q r)) s)"
    shows "htriple P c Q"
    apply (rule htripleI)
    using assms unfolding vcg_tag_defs STATE_def .

      
  lemma htriple_decompI[vcg_decomp_rules]: 
    "DECOMP_HTRIPLE (htriple P c Q) \<Longrightarrow> htriple P c Q"
    unfolding vcg_tag_defs by auto

  lemma htriple_assn_cong[htriple_vcg_intro_congs]: 
    "P\<equiv>P' \<Longrightarrow> Q\<equiv>Q' \<Longrightarrow>  htriple P c Q \<equiv> htriple P' c Q'" by simp

  lemma htriple_ent_pre:
    "P\<turnstile>P' \<Longrightarrow> htriple P' c Q \<Longrightarrow> htriple P c Q"
    unfolding entails_def
    apply (erule cons_rule) by blast+
    
  lemma htriple_ent_post:
    "\<lbrakk>\<And>r. Q r \<turnstile> Q' r\<rbrakk> \<Longrightarrow> htriple P c Q \<Longrightarrow> htriple P c Q'"
    unfolding entails_def
    apply (erule cons_rule) by blast+
   
  lemma htriple_pure_preI: "\<lbrakk>pure_part P \<Longrightarrow> htriple P c Q\<rbrakk> \<Longrightarrow> htriple P c Q"  
    using entails_pureI htriple_ent_pre by blast
    
  
  declaration \<open>
    fn phi => (Basic_VCG.add_xformer (map (Morphism.thm phi) @{thms DECOMP_HTRIPLEI},@{binding decomp_htriple_xformer},
      fn ctxt => 
        (full_simp_tac (put_simpset HOL_basic_ss ctxt 
          addsimps (Named_Theorems.get ctxt @{named_theorems vcg_tag_defs})
          |> fold Simplifier.add_cong (Named_Theorems.get ctxt @{named_theorems htriple_vcg_intro_congs})
        ))
        THEN' resolve_tac ctxt (Named_Theorems.get ctxt @{named_theorems htriple_vcg_intros})
    ))
  \<close>
  
  
end  

end
