section \<open>General Domain Theory\<close>
theory RefineG_Domain
imports Basic_Imports (*"HOL-Library.Complete_Partial_Order2"*) (*Complete_Partial_Order2*)
begin

subsection \<open>Additions to Partial Function\<close>

lemma fun_lub_empty: "fun_lub L {} = (\<lambda>x. L {})"
  unfolding fun_lub_def by simp

lemma img_lub_empty: "img_lub f g L {} = g (L {})"  
  unfolding img_lub_def by simp
  
lemma flat_lub_empty[simp]: "flat_lub b {} = b"
  unfolding flat_lub_def by auto  
  
lemma flat_lub_single[simp]: "flat_lub b {x} = x"  
  unfolding flat_lub_def by (auto)

lemma flat_lub_double[simp]: "flat_lub b {b,x} = x"
  unfolding flat_lub_def by (auto)
  
  
lemma fun_ordD: "fun_ord le f g \<Longrightarrow> le (f x) (g x)"
  by (auto simp: fun_ord_def)

lemma fun_ord_mono_alt: "monotone le (fun_ord le') f \<longleftrightarrow> (\<forall>x. monotone le le' (\<lambda>y. f y x))"
  by (metis (mono_tags, lifting) fun_ord_def monotone_def)
  
(* See thm Partial_Function_Definitions.admissible_fun. We removed an unnecessary precondition. *)
lemma admissible_fun'(*[cont_intro]*):
  fixes le :: "'a\<Rightarrow>'a\<Rightarrow>bool" and lub :: "'a set \<Rightarrow> 'a"
  assumes adm: "\<And>x. ccpo.admissible lub le (Q x)"
  shows "ccpo.admissible  (fun_lub lub) (fun_ord le) (\<lambda>f. \<forall>x. Q x (f x))"
proof (rule ccpo.admissibleI)
  fix A :: "('b \<Rightarrow> 'a) set"
  assume Q: "\<forall>f\<in>A. \<forall>x. Q x (f x)"
  assume ch: "Complete_Partial_Order.chain (fun_ord le) A"
  assume "A \<noteq> {}"
  hence non_empty: "\<And>a. {y. \<exists>f\<in>A. y = f a} \<noteq> {}" by auto
  show "\<forall>x. Q x (fun_lub lub A x)"
    unfolding fun_lub_def
    by (rule allI, rule ccpo.admissibleD[OF adm chain_fun[OF ch] non_empty])
      (auto simp: Q)
qed

(* See thm Partial_Function_Definitions.admissible_image.
  We removed pfun precondition, 
  moreover, we removed inj-precondition and changed direction of inv
*)
lemma admissible_image'(*[cont_intro]*):
  fixes le :: "'a\<Rightarrow>'a\<Rightarrow>bool" and lub :: "'a set \<Rightarrow> 'a"
  assumes adm: "ccpo.admissible lub le (P \<circ> g)"
  (*assumes inj: "\<And>x y. f x = f y \<Longrightarrow> x = y"*)
  assumes inv: "\<And>y. g (f y) = y"
  shows "ccpo.admissible (img_lub f g lub) (img_ord f le) P"
proof (rule ccpo.admissibleI)
  fix A assume "Complete_Partial_Order.chain (img_ord f le) A"
  then have ch': "Complete_Partial_Order.chain le (f ` A)"
    by (auto simp: img_ord_def intro: chainI dest: chainD)
  assume "A \<noteq> {}"
  assume P_A: "\<forall>x\<in>A. P x"
  have "(P \<circ> g) (lub (f ` A))" using adm ch'
  proof (rule ccpo.admissibleD)
    fix x assume "x \<in> f ` A"
    with P_A show "(P \<circ> g) x" by (auto simp: inv)
  qed(simp add: \<open>A \<noteq> {}\<close>)
  thus "P (img_lub f g lub A)" unfolding img_lub_def by simp
qed


lemma flat_chain_alt: "Complete_Partial_Order.chain (flat_ord b) C \<longleftrightarrow> C\<subseteq>{b} \<or> (\<exists>x. C-{b} = {x})"  
  unfolding chain_def flat_ord_def apply simp apply safe
  apply blast
  apply blast
  by (metis Diff_iff insert_iff)
  
lemma flat_chain_alt': "Complete_Partial_Order.chain (flat_ord b) C \<longleftrightarrow> C={} \<or> C={b} \<or> (\<exists>x. x\<noteq>b \<and> (C={x} \<or> C = {b,x}))"
  apply (simp add: flat_chain_alt)
  by auto

lemma flat_chain_cases: 
  assumes "Complete_Partial_Order.chain (flat_ord b) C"
  obtains (empty) "C={}" 
        | (bot) "C={b}" 
        | (single) x where "x\<noteq>b" "C={x}" 
        | (double) x where "x\<noteq>b" "C={b,x}"
  using assms unfolding flat_chain_alt' 
  by blast
  
lemma admissible_flat(*[cont_intro]*): "ccpo.admissible (flat_lub b) (flat_ord b) P"
  unfolding ccpo.admissible_def
  by (auto elim!: flat_chain_cases simp: flat_lub_def)


subsection \<open>Wf-Induction for Fixed Points\<close>


lemma wf_fixp_induct:
  \<comment> \<open>Well-Founded induction for arbitrary fixed points\<close>
  assumes fixp_unfold: "fp B = B (fp B)"
  assumes WF: "wf V"
  assumes STEP: "\<And>f t. \<lbrakk> 
    \<And>t'. \<lbrakk> (t',t)\<in>V \<rbrakk> \<Longrightarrow> P t' f; fp B = f \<rbrakk> \<Longrightarrow> P t (B f)"
  shows "P t (fp B)"
  using WF
  apply (induct t rule: wf_induct_rule)
  apply (subst fixp_unfold)
  apply (rule STEP)
  apply (simp)
  apply (simp)
  done

(*
lemma wf_fixp_induct:
  \<comment> \<open>Well-Founded induction for arbitrary fixed points\<close>
  fixes a :: 'a
  assumes fixp_unfold: "fp B = B (fp B)"
  assumes WF: "wf V"
  assumes STEP: "\<And>f x. \<lbrakk> 
    \<And>x'. \<lbrakk> (x',x)\<in>V \<rbrakk> \<Longrightarrow> P x' (f x'); fp B = f \<rbrakk> \<Longrightarrow> P x (B f x)"
  shows "P x (fp B x)"
  using WF
  apply (induct x rule: wf_induct_rule)
  apply (subst fixp_unfold)
  apply (rule STEP)
  apply (simp)
  apply (simp)
  done
*)
  

subsection \<open>Generic Flat Monad Setup\<close>
text \<open>
  Recursion setup for monads with a single bottom element
\<close>
locale monad_rec =
  fixes bot :: 'r
begin
  abbreviation "leq \<equiv> flat_ord bot"
  abbreviation "lub \<equiv> flat_lub bot"

  lemma leq_bot: "leq bot x" by (simp add: flat_ord_def)  
  
  sublocale PF: partial_function_definitions leq lub
    by (rule flat_interpretation)

  (*sublocale order_mono_setup: mono_setup_loc leq
    apply standard 
    by (simp add: PF.leq_refl)*)
    
  sublocale pf_ccpo: ccpo "PF.lub_fun" "PF.le_fun" "mk_less PF.le_fun"  
    by (rule PF.ccpo)
  
  abbreviation "mono_body \<equiv> monotone PF.le_fun PF.le_fun"        
  (*abbreviation "PF_mono \<equiv> monotone PF.le_fun leq"*)

  subsubsection \<open>Basic Theory\<close>  
  
  lemma basic_admissible_pointwise:
    shows "PF.admissible (\<lambda>a. \<forall>x. Q x (a x))"
    by (simp add: admissible_flat admissible_fun')
  
  lemma basic_pointwise_induct:
    assumes BOT: "\<forall>x. Q x bot"
    assumes MONO: "mono_body F"
    assumes STEP: "\<And>D x. \<lbrakk>\<And>x. Q x (D x)\<rbrakk> \<Longrightarrow> Q x (F D x)"
    shows "Q x (pf_ccpo.fixp F x)"
  proof -
    have "\<forall>x. Q x (pf_ccpo.fixp F x)"
      apply (rule pf_ccpo.fixp_induct[OF basic_admissible_pointwise MONO])
      subgoal by (simp add: BOT fun_lub_empty)
      subgoal using STEP by blast
      done
    thus ?thesis by blast
  qed    
  
  lemma pf_ccpo_fixp_mono: "mono_body F \<Longrightarrow> mono_body F' \<Longrightarrow> (\<And>D x. leq (F D x) (F' D x)) \<Longrightarrow> leq (pf_ccpo.fixp F x) (pf_ccpo.fixp F' x)"
    by (metis (full_types) fun_ord_def pf_ccpo.fixp_mono)
  

      
  definition "rec F x \<equiv> if mono_body F then pf_ccpo.fixp F x else bot"

  lemma rec_unfold: "mono_body F \<Longrightarrow> rec F = F (rec F)"
    unfolding rec_def by (simp flip: pf_ccpo.fixp_unfold)

  lemma basic_rec_wf_rule:  
    assumes M: "mono_body F"
    assumes WF: "wf V"
    assumes STEP: "\<And>f t. \<lbrakk>\<And>t'. (t', t) \<in> V \<Longrightarrow> P t' f; rec F = f\<rbrakk> \<Longrightarrow> P t (F f)"
    shows "P t (rec F)"
    apply (rule wf_fixp_induct[where fp=rec and P=P,OF rec_unfold[OF M]])
    by fact+
    
    
  lemma basic_rec_induct:
    assumes BOT: "P (\<lambda>_. bot)"
    assumes ADM: "PF.admissible P"
    assumes STEP: "\<And>D. \<lbrakk>P D\<rbrakk> \<Longrightarrow> P (F D)"
    shows "P (rec F)"  
    unfolding rec_def
    apply (cases "mono_body F")
    apply (all \<open>clarsimp simp: BOT\<close>)
  proof -
    assume "mono_body F"
    show "P (pf_ccpo.fixp F)"
      apply (induction rule: pf_ccpo.fixp_induct)
      subgoal by fact
      subgoal by fact
      subgoal by (simp add: BOT fun_lub_empty)
      subgoal by fact
      done
      
  qed    
    
  lemma pointwise_rec_induct:
    assumes BOT: "\<forall>x. Q x bot"
    assumes STEP: "\<And>D x. \<lbrakk>\<And>x. Q x (D x)\<rbrakk> \<Longrightarrow> Q x (F D x)"
    shows "Q x (rec F x)"  
    unfolding rec_def
    apply (clarsimp simp: BOT)
    using basic_pointwise_induct[where Q=Q,OF BOT _ STEP]
    by simp
  

  (*
    TODO: Make this proof generic, in partial_function_definitions or so.
  *)
  
  lemma REC_mono_aux:
    assumes MONO: "\<And>D. monotone PF.le_fun PF.le_fun (B D)"
    assumes 1: "monotone ord (fun_ord PF.le_fun) B"
    shows "monotone ord PF.le_fun (\<lambda>D. rec (B D))"
    unfolding rec_def apply (simp add: MONO)
    apply (rule monotoneI)
    apply (rule ccpo.fixp_lowerbound[OF PF.ccpo MONO])
    apply (rewrite in "PF.le_fun _ \<hole>" ccpo.fixp_unfold[OF PF.ccpo MONO])
    supply R=fun_ordD[of PF.le_fun "B x" "B y" for x y]
    apply (rule R)
    apply (rule monotoneD[OF 1])
    .
  
  
  lemma REC_mono[partial_function_mono]:
    assumes MONO: "\<And>D x. PF.mono_body (\<lambda>E. B D E x)"
    assumes 1: "\<And>E x. monotone ord leq (\<lambda>D. B D E x)"
    shows "monotone ord leq (\<lambda>D. rec (B D) x)"
    using assms REC_mono_aux fun_ord_mono_alt by metis
    
end

locale admissible_property = monad_rec bot for bot :: 'r +
  fixes P :: "('a \<Rightarrow> 'r) \<Rightarrow> bool"
  assumes BOT: "P (\<lambda>_. bot)"
  assumes ADM: "PF.admissible P"
begin  

  lemmas rec_induct = basic_rec_induct[of P, OF BOT ADM]

end




subsection \<open>Generic State Monad Setup\<close>
text \<open>
  We assume that the state monad is encapsulated into a datatype, 
  using the morphisms \<open>run\<close> and \<open>mkm\<close>. Moreover, the raw result 
  type \<open>'r\<close> needs a bottom element.
\<close>
locale state_monad_rec =
  fixes run :: "'m \<Rightarrow> 's \<Rightarrow> 'r"
  fixes mkm :: "('s\<Rightarrow>'r) \<Rightarrow> 'm"
  fixes bot :: 'r
  fixes bot' :: 'm
  assumes run_inv: "run (mkm f) = f"
  assumes run_injI: "run m = run m' \<Longrightarrow> m=m'"
  assumes bot'_def: "bot' = mkm (\<lambda>_. bot)"
begin

  lemma run_eq_conv: "run m = run m' \<longleftrightarrow> m=m'" using run_injI by auto
  
  lemma mkm_inv: "mkm (run m) = m"
    using run_eq_conv run_inv by blast
  
  
  abbreviation "leq \<equiv> img_ord run (fun_ord (flat_ord bot))"
  abbreviation "lub \<equiv> img_lub run mkm (fun_lub (flat_lub bot))"

  lemma leq_bot: "leq bot' x"
    by (simp add: flat_ord_def fun_ord_def img_ord_def run_inv bot'_def)  
  
  sublocale PF: partial_function_definitions leq lub
    apply (rule partial_function_image)
    apply (rule partial_function_lift)
    apply (rule flat_interpretation)
    by (auto simp: run_inv run_injI)

  (*  
  sublocale order_mono_setup: mono_setup_loc leq
    apply standard 
    by (simp add: PF.leq_refl)
  *)  
    
  sublocale pf_ccpo: ccpo "PF.lub_fun" "PF.le_fun" "mk_less PF.le_fun"  
    by (rule PF.ccpo)

    
  abbreviation "mono_body \<equiv> monotone PF.le_fun PF.le_fun"        
  (*abbreviation "PF_mono \<equiv> monotone PF.le_fun leq"*)

  (* Should anyone use mono_body. *)
  lemma mono_body_to_PF_monoI[partial_function_mono]: "\<lbrakk>\<And>x. monotone PF.le_fun leq (\<lambda>D. B D x) \<rbrakk> \<Longrightarrow> mono_body B"  
    by (simp add: fun_ord_mono_alt)
    
  
  subsection \<open>Pointwise Properties\<close>      

  text \<open>Intuition of a pointwise strict property:
    The result is tested individually for each argument and state. 
    Moreover, a bot result is always OK, i.e., in particular, the property holds for 
    the constant bot function.
    
    Naturally, refinement properties are pointwise in the abstract argument. 
    Note that bot is the most unspecific result here, i.e., refined by all other results.
  \<close>

  subsubsection \<open>Basic Theory\<close>  
  
  lemma basic_admissible_pointwise:
    shows "PF.admissible (\<lambda>a. \<forall>x s. Q x s (run (a x) s))"
    apply (rule admissible_fun')
    apply (rule admissible_image')
    apply (simp_all add: comp_def mkm_inv run_inv)    
    apply (rule admissible_fun')
    by (rule admissible_flat)
    
  
  lemma basic_pointwise_induct:
    assumes BOT: "\<forall>x s. Q x s bot"
    assumes MONO: "mono_body F"
    assumes STEP: "\<And>D x s. \<lbrakk>\<And>x s. Q x s (run (D x) s)\<rbrakk> \<Longrightarrow> Q x s (run (F D x) s)"
    shows "Q x s (run (pf_ccpo.fixp F x) s)"  
  proof -
    have "\<forall>x s. Q x s (run (pf_ccpo.fixp F x) s)"
      apply (rule pf_ccpo.fixp_induct[OF basic_admissible_pointwise MONO])
      subgoal by (simp add: fun_lub_empty img_lub_empty flat_lub_empty BOT run_inv)
      subgoal using STEP by blast
      done
    thus ?thesis by blast
  qed    
  
  lemma pf_ccpo_fixp_mono: "mono_body F \<Longrightarrow> mono_body F' \<Longrightarrow> (\<And>D x. leq (F D x) (F' D x)) \<Longrightarrow> leq (pf_ccpo.fixp F x) (pf_ccpo.fixp F' x)"
    by (metis (full_types) fun_ord_def pf_ccpo.fixp_mono)
  
  
  definition "rec F x \<equiv> if mono_body F then pf_ccpo.fixp F x else bot'"

  lemma rec_unfold: "mono_body F \<Longrightarrow> rec F = F (rec F)"
    unfolding rec_def by (simp flip: pf_ccpo.fixp_unfold)

  lemma basic_rec_wf_rule:  
    assumes M: "mono_body F"
    assumes WF: "wf V"
    assumes STEP: "\<And>f t. \<lbrakk>\<And>t'. (t', t) \<in> V \<Longrightarrow> P t' f; rec F = f\<rbrakk> \<Longrightarrow> P t (F f)"
    shows "P t (rec F)"
    apply (rule wf_fixp_induct[where fp=rec and P=P,OF rec_unfold[OF M]])
    by fact+
    
    
  lemma basic_rec_induct:
    assumes BOT: "P (\<lambda>_. bot')"
    assumes ADM: "PF.admissible P"
    assumes STEP: "\<And>D. \<lbrakk>P D\<rbrakk> \<Longrightarrow> P (F D)"
    shows "P (rec F)"  
    unfolding rec_def
    apply (cases "mono_body F")
    apply (all \<open>clarsimp simp: run_inv BOT\<close>)
  proof -
    assume "mono_body F"
    show "P (pf_ccpo.fixp F)"
      apply (induction rule: pf_ccpo.fixp_induct)
      subgoal by (simp add: comp_def ADM mkm_inv)
      subgoal by fact
      subgoal by (simp add: fun_lub_empty img_lub_empty BOT[unfolded bot'_def] run_inv)
      subgoal by fact
      done
      
  qed    
    
  lemma pointwise_rec_induct:
    assumes BOT: "\<forall>x s. Q x s bot"
    assumes STEP: "\<And>D x s. \<lbrakk>\<And>x s. Q x s (run (D x) s)\<rbrakk> \<Longrightarrow> Q x s (run (F D x) s)"
    shows "Q x s (run (rec F x) s)"  
    unfolding rec_def
    apply (clarsimp simp: run_inv BOT bot'_def)
    using basic_pointwise_induct[where Q=Q,OF BOT _ STEP]
    by simp


  (*
    TODO: Make this proof generic, in partial_function_definitions or so.
  *)
  lemma REC_mono_aux:
    assumes MONO: "\<And>D. monotone PF.le_fun PF.le_fun (B D)"
    assumes 1: "monotone ord (fun_ord PF.le_fun) B"
    shows "monotone ord PF.le_fun (\<lambda>D. rec (B D))"
    unfolding rec_def apply (simp add: MONO)
    apply (rule monotoneI)
    apply (rule ccpo.fixp_lowerbound[OF PF.ccpo MONO])
    apply (rewrite in "PF.le_fun _ \<hole>" ccpo.fixp_unfold[OF PF.ccpo MONO])
    supply R=fun_ordD[of PF.le_fun "B x" "B y" for x y]
    apply (rule R)
    apply (rule monotoneD[OF 1])
    .

  lemma REC_mono[partial_function_mono]:
    assumes MONO: "\<And>D x. PF.mono_body (\<lambda>E. B D E x)"
    assumes 1: "\<And>E x. monotone ord leq (\<lambda>D. B D E x)"
    shows "monotone ord leq (\<lambda>D. rec (B D) x)"
    using assms REC_mono_aux fun_ord_mono_alt by metis
    
end    
    


locale state_admissible_property = state_monad_rec run mkm bot bot' for run :: "'m \<Rightarrow> 's \<Rightarrow> 'r" and mkm bot bot' +
  fixes P :: "('a \<Rightarrow> 'm) \<Rightarrow> bool"
  assumes BOT: "P (\<lambda>_. bot')"
  assumes ADM: "PF.admissible P"
begin  

  lemmas rec_induct = basic_rec_induct[of P, OF BOT ADM]

end



end

